# coding: utf-8
"""

"""
import copy
import csv
import dataclasses
import decimal
import difflib
import io
import json
import base64
import functools
import hashlib
import os
import re
import typing
from io import BytesIO
from itertools import zip_longest
from urllib.parse import quote_plus
from datetime import datetime, timezone
from math import log10, floor

import flask
import flask_babel
import jinja2.filters
import werkzeug
from flask_babel import format_datetime, format_date, get_locale
from flask_login import current_user
from babel import numbers
import markupsafe
import qrcode
import qrcode.image.svg
import plotly
import pytz
import numpy as np

from ..logic import errors
from ..logic.caching import cache_per_request
from ..logic.components import get_component_or_none, get_component_id_by_uuid, get_component_by_uuid, Component
from ..logic.datatypes import Quantity
from ..logic.eln_import import get_eln_import_for_object, get_import_signed_by
from ..logic.errors import UserIsReadonlyError
from ..logic.units import prettify_units
from ..logic.notifications import get_num_notifications
from ..logic.markdown_to_html import markdown_to_safe_html
from ..logic.users import get_user, User, get_user_by_federated_user
from ..logic.utils import get_translated_text, get_all_translated_texts, show_numeric_tags_warning, relative_url_for
from ..logic.schemas.conditions import are_conditions_fulfilled
from ..logic.schemas.data_diffs import DataDiff, apply_diff, invert_diff
from ..logic.schemas.utils import get_property_paths_for_schema
from ..logic.schemas import get_default_data
from ..logic.actions import Action
from ..logic.action_types import ActionType
from ..logic.info_pages import InfoPage, get_info_pages_for_endpoint
from ..logic.instruments import Instrument
from ..logic.instrument_log_entries import InstrumentLogFileAttachment
from ..logic.action_permissions import get_sorted_actions_for_user, get_actions_with_permissions
from ..logic.languages import get_user_language
from ..logic.locations import Location, LocationType, get_location, get_unhandled_object_responsibility_assignments, is_full_location_tree_hidden, get_locations_tree
from ..logic.location_permissions import get_user_location_permissions, get_locations_with_user_permissions
from ..logic.datatypes import JSONEncoder
from ..logic.security_tokens import generate_token
from ..logic.object_permissions import get_user_object_permissions
from ..logic.objects import get_object
from ..logic.groups import Group, get_groups
from ..logic.projects import Project, get_projects, get_child_project_ids, get_parent_project_ids, get_project
from ..logic.group_categories import get_group_category_tree, get_group_categories, get_basic_group_categories, get_project_group_categories, get_full_group_category_name, GroupCategoryTree
from ..logic.files import File, get_file as get_file_logic
from ..models import Permissions, Object
from ..utils import generate_content_security_policy_nonce
from .info_pages import InfoPageAcknowledgementForm


class JinjaFilter:
    JinjaFilterT = typing.TypeVar('JinjaFilterT')
    filters: typing.Dict[str, typing.Any] = {}

    def __init__(self, name: str = '') -> None:
        self._name = name

    def __call__(self, func: JinjaFilterT) -> JinjaFilterT:
        if not self._name:
            self._name = func.__name__  # type: ignore[attr-defined]
        self.filters[self._name] = func
        return func


class JinjaFunction:
    JinjaFunctionT = typing.TypeVar('JinjaFunctionT')
    functions: typing.Dict[str, typing.Any] = {}

    def __init__(self, name: str = '') -> None:
        self._name = name

    def __call__(self, func: JinjaFunctionT) -> JinjaFunctionT:
        if not self._name:
            self._name = func.__name__  # type: ignore[attr-defined]
        self.functions[self._name] = func
        return func


JinjaFilter()(hash)
JinjaFilter()(prettify_units)
JinjaFilter('urlencode')(quote_plus)
JinjaFilter()(markdown_to_safe_html)
JinjaFilter()(get_translated_text)
JinjaFilter()(get_all_translated_texts)
JinjaFilter()(bool)

JinjaFunction()(get_component_or_none)
JinjaFunction()(get_component_id_by_uuid)
JinjaFunction()(get_unhandled_object_responsibility_assignments)
JinjaFunction()(is_full_location_tree_hidden)
JinjaFunction()(get_eln_import_for_object)
JinjaFunction()(relative_url_for)
JinjaFunction()(zip_longest)
JinjaFunction()(get_default_data)
JinjaFunction()(apply_diff)
JinjaFunction()(invert_diff)
JinjaFunction()(get_import_signed_by)
JinjaFunction()(InfoPageAcknowledgementForm)

qrcode_cache: typing.Dict[str, str] = {}


def generate_qrcode(url: str, should_cache: bool = True) -> str:
    """
    Generate a QR code (as data URI) to a given URL.

    :param url: the url the QR code should reference
    :param should_cache: whether or not the QR code should be cached
    :return: a data URI to a base64 encoded SVG image
    """
    if should_cache and url in qrcode_cache:
        return qrcode_cache[url]
    image = qrcode.make(url, image_factory=qrcode.image.svg.SvgPathFillImage)
    image_stream = BytesIO()
    image.save(image_stream)
    image_stream.seek(0)
    qrcode_url = 'data:image/svg+xml;base64,' + base64.b64encode(image_stream.read()).decode('utf-8')
    if should_cache:
        qrcode_cache[url] = qrcode_url
    return qrcode_url


@JinjaFunction()
def get_file(
        file_id: int,
        object_id: int,
        component_uuid: typing.Optional[str] = None
) -> typing.Optional[File]:
    component_id = get_component_id_by_uuid(component_uuid)
    try:
        return get_file_logic(file_id, object_id, component_id)
    except errors.FileDoesNotExistError:
        return None


@JinjaFilter()
def has_preview(file: File) -> bool:
    if file.storage not in {'database', 'federation'}:
        return False
    if file.preview_image_binary_data and file.preview_image_mime_type:
        return True
    file_name = file.original_file_name
    file_extension = os.path.splitext(file_name)[1]
    return file_extension.lower() in flask.current_app.config.get('MIME_TYPES', {})


@JinjaFilter()
def has_preview_image(file: File) -> bool:
    if not file.preview_image_mime_type or not file.preview_image_binary_data:
        return is_image(file)
    mime_types: typing.Dict[str, str] = flask.current_app.config.get('MIME_TYPES', {})
    mime_type = file.preview_image_mime_type.lower()
    return mime_type in mime_types.values() and mime_type.startswith('image/')


def file_name_is_image(file_name: str) -> bool:
    file_extension = os.path.splitext(file_name)[1]
    mime_types: typing.Dict[str, str] = flask.current_app.config.get('MIME_TYPES', {})
    return mime_types.get(file_extension.lower(), '').startswith('image/')


@JinjaFilter()
def is_image(file: File) -> bool:
    # federation files are not recognized as images to prevent loading their thumbnails from other database
    if file.storage != 'database':
        return False
    return file_name_is_image(file.original_file_name)


@JinjaFilter()
def attachment_is_image(file_attachment: InstrumentLogFileAttachment) -> bool:
    return file_name_is_image(file_attachment.file_name)


@JinjaFilter()
def get_num_unread_notifications(user: User) -> int:
    return get_num_notifications(user.id, unread_only=True)


def check_current_user_is_not_readonly() -> None:
    if current_user.is_readonly:
        raise UserIsReadonlyError()


@JinjaFilter('plot')
def plotly_base64_image_from_json(object: typing.Dict[str, typing.Any]) -> typing.Optional[str]:
    try:
        fig_plot = plotly.io.from_json(json.dumps(object))
    except ValueError:
        return None
    image_stream = BytesIO()
    fig_plot.write_image(image_stream, "svg")
    image_stream.seek(0)
    return 'data:image/svg+xml;base64,' + base64.b64encode(image_stream.read()).decode('utf-8')


@JinjaFilter()
def plotly_chart_get_title(plotly_object: typing.Dict[str, typing.Any]) -> str:
    layout = plotly_object.get('layout')
    if isinstance(layout, dict):
        title = layout.get('title')
        if isinstance(title, str):
            return title
        if isinstance(title, dict):
            text = title.get('text')
            if isinstance(text, str):
                return text
    return ""


@JinjaFilter()
def to_json_no_extra_escapes(
        json_object: typing.Dict[str, typing.Any],
        indent: typing.Optional[int] = None
) -> str:
    return json.dumps(json_object, indent=indent)


@JinjaFilter('generic_format_datetime')
def generic_format_datetime(
        utc_datetime: typing.Union[str, datetime]
) -> str:
    """
    Returns a datetime in the common format and in UTC.

    :param utc_datetime: a datetime str or object in UTC
    :return: the formatted datetime
    """
    if isinstance(utc_datetime, str):
        # assume the datetime is already formatted
        return utc_datetime
    return utc_datetime.strftime('%Y-%m-%d %H:%M:%S')


@JinjaFilter('babel_format_datetime')
def custom_format_datetime(
        utc_datetime: typing.Union[str, datetime],
        format: typing.Optional[str] = None
) -> typing.Union[str, datetime]:
    """
    Returns a reformatted date in the given format.

    :param utc_datetime: a datetime string or object in UTC
    :param format: the target format or None for the user's medium-length format
    :return: the reformatted datetime or utc_datetime in case of an error
    """
    try:
        if not isinstance(utc_datetime, datetime):
            utc_datetime = parse_datetime_string(utc_datetime)
        if format is None:
            format2 = 'medium'
            return format_datetime(utc_datetime, format=format2)
        else:
            if utc_datetime.tzinfo is None:
                utc_datetime = pytz.utc.localize(utc_datetime)
            local_datetime = utc_datetime.astimezone(pytz.timezone(current_user.timezone or 'UTC'))
            return local_datetime.strftime(format)
    except ValueError:
        return utc_datetime


@JinjaFilter('babel_format_date')
def custom_format_date(
        utc_datetime: typing.Union[datetime, str],
        format: typing.Optional[str] = None
) -> str:
    """
    Return a formatted date.

    :param utc_datetime: a datetime string or object in UTC
    :param format: the babel date format, or None
    :return: the formatted date
    """
    if isinstance(utc_datetime, str):
        utc_datetime = datetime.strptime(utc_datetime, '%Y-%m-%d %H:%M:%S')
    return format_date(utc_datetime, format=format)


@JinjaFilter('babel_format_time')
def custom_format_time(
        utc_datetime: typing.Union[str, datetime],
        format: typing.Optional[str] = None
) -> typing.Union[str, datetime]:
    """
    Return a formatted time.

    :param utc_datetime: a datetime string or object in UTC
    :param format: the babel time format, or None
    :return: the formatted time or utc_datetime in case of an error
    """
    try:
        if not isinstance(utc_datetime, datetime):
            utc_datetime = parse_datetime_string(utc_datetime)
        return flask_babel.format_time(utc_datetime, format=format)
    except ValueError:
        return utc_datetime


@JinjaFilter('format_int')
def format_int(decimal: int) -> str:
    locale = get_locale()
    return typing.cast(str, numbers.format_decimal(decimal, locale=locale))


@JinjaFilter('babel_format_number')
def custom_format_number(
    number: typing.Union[str, int, float, decimal.Decimal],
    display_digits: typing.Optional[int] = None,
    integral_digits: typing.Optional[int] = None,
    disable_scientific_format: bool = False
) -> str:
    """
    Return the formatted number.

    :param number: either an int or a float, or a string representation of either
    :param display_digits: number of decimals to use, or None
    :param integral_digits: minimal number of digits in the integral part, or None
    :param disable_scientific_format: disables output in scientific format
    :return: the formatted number
    """
    try:
        # if number is a string that can not be formatted. Wrong inputs...
        number = float(number)
    except ValueError:
        return typing.cast(str, number)
    if type(display_digits) is not int:
        display_digits = None
    else:
        # there cannot be a negative number of digits
        if display_digits < 0:
            display_digits = 0

        # explicitly round the number, as that might change the exponent
        number = round(number, display_digits)

    locale = get_locale()
    if number == 0:
        exponent = 0
    else:
        exponent = int(floor(log10(abs(number))))
    if not disable_scientific_format:
        # for very small or very large absolute numbers, the exponential format should be used
        use_exponential_format = not -5 < exponent < 6
    else:
        exponent = 0
        use_exponential_format = False

    if integral_digits is not None:
        positive_format = '0' * integral_digits
    else:
        positive_format = '0'
    if display_digits is not None:
        if use_exponential_format:
            display_digits += exponent
            if display_digits > 15:
                display_digits = 15
        else:
            internal_integral_digits = max(0, exponent + 1)
            if display_digits + internal_integral_digits > 15:
                display_digits = 15 - internal_integral_digits
            if display_digits < 0:
                display_digits = 0

        positive_format = positive_format + '.' + '0' * display_digits
    else:
        positive_format = positive_format + '.###'
    if use_exponential_format:
        # including E will enable exponential format, 0 means the exponent should be shown even if 0
        positive_format += 'E0'
    negative_format = '-' + positive_format
    f = positive_format + ';' + negative_format

    if use_exponential_format:
        return numbers.format_scientific(
            number,
            locale=locale,
            format=f,
            decimal_quantization=False
        )
    else:
        # babel.numbers doesn't explicitly export decimal, but it is
        # recommended to use babel.numbers.decimal instead of the standard
        # library module, in case babel decides to use a different decimal
        # implementation in the future
        with numbers.decimal.localcontext() as ctx:  # type: ignore[attr-defined]
            ctx.prec = 15
            return typing.cast(str, numbers.format_decimal(
                numbers.decimal.Decimal(number),  # type: ignore[attr-defined]
                locale=locale,
                format=f,
                decimal_quantization=False
            ))


@JinjaFilter('format_time')
def format_time(
    magnitude_in_base_units: float,
    units: str,
    display_digits: typing.Optional[int] = None
) -> str:
    if units not in {'min', 'h'}:
        raise errors.MismatchedUnitError()
    decimal = 0.0
    magnitude_str = str(magnitude_in_base_units)
    if '.' in magnitude_str:
        # string manipulation to prevent ugly float-arithmetics
        integral_str, decimal_str = magnitude_str.split('.', maxsplit=1)
        magnitude_in_base_units = int(integral_str)
        decimal = float(f'0.{decimal_str}')
    seconds = magnitude_in_base_units % 60
    magnitude_in_base_units -= seconds
    seconds += decimal
    if units == 'h':
        minutes = int(magnitude_in_base_units // 60) % 60
        magnitude_in_base_units -= minutes * 60
        hours = int(magnitude_in_base_units // 3600)
        return f'{hours:02d}:{minutes:02d}:{custom_format_number(seconds, display_digits, 2, True)}'
    if units == 'min':
        minutes = int(magnitude_in_base_units // 60)
        return f'{minutes:02d}:{custom_format_number(seconds, display_digits, 2, True)}'
    # fall back to regular quantity formatting
    return custom_format_quantity(
        data={
            'magnitude_in_base_units': magnitude_in_base_units,
            'units': units
        },
        schema={
            'display_digits': display_digits
        }
    )


@JinjaFilter('format_quantity')
def custom_format_quantity(
        data: typing.Optional[typing.Dict[str, typing.Any]],
        schema: typing.Dict[str, typing.Any]
) -> str:
    if data is None:
        mdash = '\u2014'
        return mdash
    if data.get('units', '1') in {'1', ''}:
        return custom_format_number(data.get('magnitude_in_base_units', 0), schema.get('display_digits', None))
    narrow_non_breaking_space = '\u202f'
    magnitude = data.get('magnitude_in_base_units', 0)
    units = data.get('units')
    if units in {'h', 'min'}:
        return f'{format_time(magnitude, units, schema.get("display_digits"))}{narrow_non_breaking_space}{data.get("units")}'
    quantity = Quantity.from_json(data)
    return custom_format_number(quantity.magnitude, schema.get('display_digits', None)) + narrow_non_breaking_space + prettify_units(quantity.units)


@JinjaFilter()
def parse_datetime_string(datetime_string: str) -> datetime:
    return datetime.strptime(datetime_string, '%Y-%m-%d %H:%M:%S').replace(tzinfo=timezone.utc)


@JinjaFilter()
def default_format_datetime(utc_datetime: typing.Union[str, datetime]) -> typing.Union[str, datetime]:
    return custom_format_datetime(utc_datetime, format='%Y-%m-%d %H:%M:%S')


@JinjaFilter()
def convert_datetime_input(datetime_input: str) -> str:
    if not datetime_input:
        return ''
    try:
        user_language = get_user_language(current_user)
        local_datetime = datetime.strptime(datetime_input, user_language.datetime_format_datetime)
        return local_datetime.strftime('%Y-%m-%d %H:%M:%S')
    except Exception:
        return ''


@JinjaFilter()
def base64encode(value: typing.Any) -> str:
    return base64.b64encode(json.dumps(value, separators=(',', ':')).encode('utf8')).decode('ascii')


@JinjaFilter('are_conditions_fulfilled')
def filter_are_conditions_fulfilled(data: typing.Dict[str, typing.Any], property_schema: typing.Dict[str, typing.Any]) -> bool:
    if not isinstance(property_schema, dict):
        return False
    if not data and 'conditions' in property_schema:
        return False
    return are_conditions_fulfilled(property_schema.get('conditions'), data)


@typing.overload
def to_string_if_dict(data: typing.Dict[str, typing.Any]) -> str:
    ...


@typing.overload
def to_string_if_dict(data: typing.Any) -> typing.Any:
    ...


@JinjaFilter()
def to_string_if_dict(data: typing.Any) -> typing.Any:
    if isinstance(data, dict):
        return str(data)
    else:
        return data


@JinjaFilter()
def get_location_name(
        location_or_location_id: typing.Union[int, Location],
        include_id: bool = False,
        language_code: typing.Optional[str] = None,
        has_read_permissions: typing.Optional[bool] = None
) -> str:
    location: typing.Optional[Location]
    location_id: int
    if type(location_or_location_id) is int:
        location_id = location_or_location_id
        try:
            location = get_location(location_id)
        except errors.LocationDoesNotExistError:
            location = None
    elif isinstance(location_or_location_id, Location):
        location = location_or_location_id
        location_id = location.id
    else:
        return flask_babel.gettext("Unknown Location")

    if location is not None and has_read_permissions is None:
        has_read_permissions = Permissions.READ in get_user_location_permissions(location_id, current_user.id)

    if location is None or not has_read_permissions:
        # location ID is always included when the location cannot be accessed
        location_name = flask_babel.gettext("Location") + f' #{location_id}'
    else:
        location_name = get_translated_text(
            location.name,
            language_code=language_code,
            default=flask_babel.gettext('Unnamed Location')
        )
        if include_id:
            location_name += f' (#{location_id})'
    return location_name


@JinjaFilter()
def get_full_location_name(
        location_or_location_id: typing.Union[int, Location],
        include_id: bool = False,
        language_code: typing.Optional[str] = None
) -> str:
    location: typing.Optional[Location]
    if type(location_or_location_id) is int:
        location_id: int = location_or_location_id
        try:
            location = get_location(location_id)
        except errors.LocationDoesNotExistError:
            location = None
    elif isinstance(location_or_location_id, Location):
        location = location_or_location_id
    else:
        location = None
    if location is None:
        return flask_babel.gettext("Unknown Location")

    full_location_name = get_location_name(location, include_id=include_id, language_code=language_code)
    while location.parent_location_id is not None:
        location = get_location(location.parent_location_id)
        full_location_name = get_location_name(location, include_id=False, language_code=language_code) + ' / ' + full_location_name
    return full_location_name


@JinjaFilter()
def to_datatype(obj: typing.Any) -> typing.Any:
    return json.loads(json.dumps(obj), object_hook=JSONEncoder.object_hook)


def get_style_aliases(style: str) -> typing.List[str]:
    return {
        'choice': ['list', 'choice'],
        'horizontal_table': ['table', 'horizontal_table'],
        'full_width_table': ['table']
    }.get(style, [style])


@JinjaFunction()
def get_style_variant(
        style: typing.Optional[typing.Union[str, typing.Dict[str, str]]],
        template_mode: str
) -> typing.Optional[str]:
    if isinstance(style, str):
        return style
    if isinstance(style, dict):
        if template_mode not in style and template_mode == 'inline_edit':
            template_mode = 'view'
        return style.get(template_mode)
    return None


def get_template(template_mode: str, default_prefix: str, schema: typing.Dict[str, typing.Any], container_style: typing.Optional[str]) -> str:
    return get_template_impl(
        template_mode=template_mode,
        default_prefix=default_prefix,
        schema_type=schema['type'],
        schema_style=get_style_variant(schema.get('style'), template_mode),
        container_style=get_style_variant(container_style, template_mode)
    )


@functools.cache
def get_template_impl(template_mode: str, default_prefix: str, schema_type: str, schema_style: typing.Optional[str], container_style: typing.Optional[str]) -> str:
    template_folder = {
        'view': 'objects/view/',
        'inline_edit': 'objects/inline_edit/',
        'form': 'objects/forms/'
    }.get(template_mode, 'objects/view/')
    system_path = os.path.join(os.path.dirname(__file__), 'templates', template_folder)
    base_file = str(schema_type) + ".html"

    file_order = [default_prefix + base_file]
    if container_style:
        for container_style_alias in get_style_aliases(container_style):
            file_order.insert(0, (default_prefix + container_style_alias + "_" + base_file))
    if schema_style:
        for style in get_style_aliases(schema_style):
            file_order.insert(0, (default_prefix + style + "_" + base_file))
    if container_style and schema_style:
        for style in get_style_aliases(schema_style):
            for container_style_alias in get_style_aliases(container_style):
                file_order.insert(0, (default_prefix + container_style_alias + "_" + style + "_" + base_file))

    for file in file_order:
        if os.path.exists(os.path.join(system_path, file)):
            return template_folder + file

    return template_folder + default_prefix + base_file


def get_property_template(template_mode: str, schema: typing.Dict[str, typing.Any], container_style: typing.Optional[typing.Union[str, typing.Dict[str, str]]]) -> str:
    return get_property_template_impl(
        template_mode=template_mode,
        schema_style=get_style_variant(schema.get('style'), template_mode),
        container_style=get_style_variant(container_style, template_mode)
    )


@functools.cache
def get_property_template_impl(template_mode: str, schema_style: typing.Optional[str], container_style: typing.Optional[str]) -> str:
    template_folder = {
        'view': 'objects/view/',
        'inline_edit': 'objects/inline_edit/',
        'form': 'objects/forms/'
    }.get(template_mode, 'objects/view/')
    system_path = os.path.join(os.path.dirname(__file__), 'templates', template_folder)

    file_order = ['regular_property.html']
    if container_style:
        for container_style_alias in get_style_aliases(container_style):
            file_order.insert(0, container_style_alias + '_property.html')
    if schema_style:
        for style in get_style_aliases(schema_style):
            file_order.insert(0, style + '_property.html')

    for file in file_order:
        if os.path.exists(os.path.join(system_path, file)):
            return os.path.join(template_folder, file)

    return os.path.join(template_folder + 'regular_property.html')


@JinjaFunction()
def get_form_property_template(schema: typing.Dict[str, typing.Any], container_style: typing.Optional[str]) -> str:
    return get_property_template('form', schema, container_style)


@JinjaFunction()
def get_view_property_template(schema: typing.Dict[str, typing.Any], container_style: typing.Optional[str]) -> str:
    return get_property_template('view', schema, container_style)


@JinjaFunction()
def get_inline_edit_property_template(schema: typing.Dict[str, typing.Any], container_style: typing.Optional[str]) -> str:
    return get_property_template('inline_edit', schema, container_style)


@JinjaFunction()
def get_form_template(schema: typing.Dict[str, typing.Any], container_style: typing.Optional[str]) -> str:
    return get_template('form', 'form_', schema, container_style)


@JinjaFunction()
def get_view_template(schema: typing.Dict[str, typing.Any], container_style: typing.Optional[str]) -> str:
    return get_template('view', '', schema, container_style)


@JinjaFunction()
def get_inline_edit_template(schema: typing.Dict[str, typing.Any], container_style: typing.Optional[str]) -> str:
    return get_template('inline_edit', 'inline_edit_', schema, container_style)


@JinjaFunction()
def get_local_month_names() -> typing.List[str]:
    locale = flask_babel.get_locale()
    if locale is None:
        return [
            "January",
            "February",
            "March",
            "April",
            "May",
            "June",
            "July",
            "August",
            "September",
            "October",
            "November",
            "December"
        ]
    else:
        return [
            locale.months['format']['wide'][i]
            for i in range(1, 13)
        ]


@JinjaFunction()
def get_templates(user_id: int) -> typing.List[Action]:
    return [
        action
        for action in get_sorted_actions_for_user(user_id=user_id)
        if action.type is not None and action.type.is_template and action.schema
    ]


@JinjaFunction()
def get_local_decimal_delimiter() -> str:
    return typing.cast(str, numbers.format_decimal(1.2346, locale=flask_babel.get_locale())[1:2])


@JinjaFunction()
def get_user_if_exists(user_id: int, component_id: typing.Optional[int] = None) -> typing.Optional[User]:
    try:
        return get_user(user_id, component_id)
    except errors.UserDoesNotExistError:
        return None
    except errors.ComponentDoesNotExistError:
        return None


@JinjaFilter()
@functools.lru_cache(maxsize=None)
def get_hash(text: str) -> str:
    return hashlib.sha256(text.encode('utf-8'), usedforsecurity=False).hexdigest()


@functools.lru_cache(maxsize=None)
def _get_fingerprint(file_path: str) -> str:
    """
    Calculate a fingerprint for a given file.

    :param file_path: path to the file that should be fingerprinted
    :return: the file fingerprint, or an empty string
    """
    try:
        block_size = 65536
        hash_method = hashlib.sha256()
        with open(file_path, 'rb') as input_file:
            buf = input_file.read(block_size)
            while buf:
                hash_method.update(buf)
                buf = input_file.read(block_size)
        return 'sha256-' + base64.b64encode(hash_method.digest()).decode('utf-8')
    except Exception:
        # if the file cannot be hashed for any reason, return an empty fingerprint
        return ''


STATIC_DIRECTORY = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'static'))


@JinjaFunction()
def fingerprinted_static(filename: str) -> str:
    fingerprint = _get_fingerprint(os.path.join(STATIC_DIRECTORY, filename))
    url = flask.url_for(
        'static',
        filename=filename,
        v=fingerprint or None
    )
    file_extension = os.path.splitext(filename)[1]
    result = markupsafe.escape(url)
    # add integrity attribute for script files and stylesheets for SRI
    if file_extension in ('.js', '.css') and fingerprint and not flask.current_app.config['DEBUG']:
        result = result + markupsafe.Markup('" integrity="' + fingerprint)
    # add a nonce for script files for CSP
    if file_extension == '.js':
        nonce = generate_content_security_policy_nonce()
        result = result + markupsafe.Markup('" nonce="' + nonce)
    return result


class SearchPathInfo(typing.TypedDict):
    types: typing.List[str]
    titles: typing.List[str]


def get_search_paths(
        actions: typing.List[Action],
        action_types: typing.List[ActionType],
        path_depth_limit: typing.Optional[int] = None,
        valid_property_types: typing.Sequence[str] = (
            'text',
            'bool',
            'quantity',
            'datetime',
            'user',
            'object_reference',
            'sample',
            'measurement',
            'tags',
        ),
        include_file_name: bool = False
) -> typing.Tuple[typing.Dict[str, SearchPathInfo], typing.Dict[typing.Optional[int], typing.Dict[str, SearchPathInfo]], typing.Dict[typing.Optional[int], typing.Dict[str, SearchPathInfo]]]:
    search_paths: typing.Dict[str, SearchPathInfo] = {}
    search_paths_by_action: typing.Dict[typing.Optional[int], typing.Dict[str, SearchPathInfo]] = {}
    search_paths_by_action_type: typing.Dict[typing.Optional[int], typing.Dict[str, SearchPathInfo]] = {}
    for action_type in action_types:
        search_paths_by_action_type[action_type.id] = {}
    for action in actions:
        if not action.schema:
            continue
        search_paths_by_action[action.id] = {}
        if action.type_id not in search_paths_by_action_type:
            search_paths_by_action_type[action.type_id] = {}
        property_infos = []
        for property_path, property_info in get_property_paths_for_schema(
                schema=action.schema,
                valid_property_types=set(valid_property_types),
                path_depth_limit=path_depth_limit
        ).items():
            property_path = '.'.join(
                key if key is not None else '?'
                for key in property_path
            )
            property_type = str(property_info.get('type', ''))
            property_title = markupsafe.escape(get_translated_text(property_info.get('title')))
            if property_type in {'object_reference', 'sample', 'measurement'}:
                # unify object_reference, sample and measurement
                property_type = 'object_reference'
            property_infos.append((property_path, property_type, property_title))
        if include_file_name and action.type and action.type.enable_files and 'text' in valid_property_types:
            property_infos.append(('file_name', 'text', markupsafe.escape('File Name')))
        for property_path, property_type, property_title in property_infos:
            search_paths_by_action[action.id][property_path] = SearchPathInfo(
                types=[property_type],
                titles=[property_title]
            )
            if property_path not in search_paths_by_action_type[action.type_id]:
                search_paths_by_action_type[action.type_id][property_path] = SearchPathInfo(
                    types=[property_type],
                    titles=[property_title]
                )
            else:
                if property_title not in search_paths_by_action_type[action.type_id][property_path]['titles']:
                    search_paths_by_action_type[action.type_id][property_path]['titles'].append(property_title)
                if property_type not in search_paths_by_action_type[action.type_id][property_path]['types']:
                    search_paths_by_action_type[action.type_id][property_path]['types'].append(property_type)
            if property_path not in search_paths:
                search_paths[property_path] = SearchPathInfo(
                    types=[property_type],
                    titles=[property_title]
                )
            else:
                if property_title not in search_paths[property_path]['titles']:
                    search_paths[property_path]['titles'].append(property_title)
                if property_type not in search_paths[property_path]['types']:
                    search_paths[property_path]['types'].append(property_type)
    return search_paths, search_paths_by_action, search_paths_by_action_type


@JinjaFunction()
def get_num_deprecation_warnings() -> int:
    return sum([
        show_numeric_tags_warning(),
    ])


@JinjaFunction()
def get_search_url(
        property_path: typing.Tuple[str, ...],
        data: typing.Optional[typing.Union[typing.List[typing.Any], typing.Dict[str, typing.Any]]],
        metadata_language: typing.Optional[str] = None
) -> typing.Optional[str]:
    search_query = get_search_query(property_path, data, metadata_language=metadata_language)
    if search_query is None:
        return None
    return flask.url_for(
        '.objects',
        q=search_query,
        advanced='on'
    )


def get_search_query(
        property_path: typing.Tuple[str, ...],
        data: typing.Optional[typing.Union[typing.List[typing.Any], typing.Dict[str, typing.Any]]],
        metadata_language: typing.Optional[str] = None
) -> typing.Optional[str]:
    attribute = '.'.join(
        str(path_element)
        for path_element in property_path
    )
    if data is None:
        return f'{attribute} == null'
    if isinstance(data, list):
        return None
    if data['_type'] in ('object', 'hazards'):
        return None
    if data['_type'] == 'bool':
        if data['value']:
            return f'{attribute} == True'
        else:
            return f'{attribute} == False'
    if data['_type'] == 'datetime':
        utc_datetime = datetime.strptime(data["utc_datetime"], '%Y-%m-%d %H:%M:%S').replace(tzinfo=timezone.utc)
        local_timezone = pytz.timezone(current_user.timezone if hasattr(current_user, 'timezone') and current_user.timezone is not None else 'UTC')
        local_datetime = utc_datetime.astimezone(local_timezone)
        local_date = local_datetime.date()
        return f'{attribute} == {local_date.year}-{local_date.month:02d}-{local_date.day:02d}'
    if data['_type'] == 'user':
        return f'{attribute} == #{data["user_id"]}'
    if data['_type'] in ('object_reference', 'sample', 'measurement'):
        return f'{attribute} == #{data["object_id"]}'
    if data['_type'] == 'quantity':
        if data['units'] == '1':
            return f'{attribute} == {data["magnitude_in_base_units"]}'
        else:
            return f'{attribute} == {to_datatype(data).magnitude}{data["units"]}'
    if data['_type'] == 'text':
        if data['text']:
            return f'{attribute} == "{get_translated_text(data["text"], metadata_language)}"'
        else:
            return f'{attribute} == ""'
    if data['_type'] == 'tags':
        return '#' + ' and #'.join(data['tags'])
    if data['_type'] == 'plotly_chart':
        title = jinja2.filters.do_striptags(plotly_chart_get_title(data['plotly']))
        if title:
            return f'{attribute} == "{str(title)}"'
        else:
            return f'{attribute} == ""'
    # fallback: find all objects that have this attribute set
    return f'!({attribute} == null)'


@JinjaFunction()
def get_table_search_url(
        property_path: typing.Tuple[str, ...],
        schema: typing.Dict[str, typing.Any]
) -> typing.Optional[str]:
    search_query = get_table_search_query(property_path, schema)
    if search_query is None:
        return None
    return flask.url_for(
        '.objects',
        q=search_query,
        advanced='on'
    )


def get_table_search_query(
        property_path: typing.Tuple[str, ...],
        schema: typing.Dict[str, typing.Any]
) -> typing.Optional[str]:
    attribute = '.'.join(
        str(path_element)
        for path_element in property_path
    )
    if schema.get('type') == 'text':
        return f'{attribute} == ""'
    if schema.get('type') == 'quantity':
        if isinstance(schema['units'], str) and schema['units'].strip() != '1':
            return f'{attribute} == 0{schema.get("units")}'
        if isinstance(schema['units'], list) and len(schema['units']) == 1 and schema['units'][0].strip() != '1':
            return f'{attribute} == 0{schema.get("units")}'
        return f'{attribute} == 0'
    if schema.get('type') == 'bool':
        return f'{attribute} == True'
    # fallback: find all objects that have this attribute set
    return f'!({attribute} == null)'


@JinjaFunction()
def build_object_location_assignment_confirmation_url(object_location_assignment_id: int) -> str:
    confirmation_url = flask.url_for(
        'frontend.accept_responsibility_for_object',
        t=generate_token(
            object_location_assignment_id,
            salt='confirm_responsibility',
            secret_key=flask.current_app.config['SECRET_KEY']
        ),
        _external=True
    )
    return confirmation_url


@JinjaFunction()
def build_object_location_assignment_declination_url(object_location_assignment_id: int) -> str:
    confirmation_url = flask.url_for(
        'frontend.decline_responsibility_for_object',
        t=generate_token(
            object_location_assignment_id,
            salt='decline_responsibility',
            secret_key=flask.current_app.config['SECRET_KEY']
        ),
        _external=True
    )
    return confirmation_url


@JinjaFunction()
def get_object_name_if_user_has_permissions(object_id: int) -> str:
    fallback_name = f"{flask_babel.gettext('Object')} #{object_id}"
    try:
        permissions = get_user_object_permissions(object_id, current_user.id)
        if Permissions.READ in permissions:
            object = get_object(object_id)
            if object.data:
                object_name = get_translated_text(object.data['name']['text'])
                return f"{object_name} (#{object_id})"
    except errors.ObjectDoesNotExistError:
        pass
    return fallback_name


def validate_orcid(orcid: str) -> typing.Tuple[bool, typing.Optional[str]]:
    # accept full ORCID iDs
    orcid_prefix = 'https://orcid.org/'
    if orcid.startswith(orcid_prefix):
        orcid = orcid[len(orcid_prefix):]
    # check ORCID iD syntax
    if not re.fullmatch(r'\d{4}-\d{4}-\d{4}-\d{3}[\dX]', orcid, flags=re.ASCII):
        return False, None
    # check ISO 7064 Mod 11, 2 checksum
    checksum = '0123456789X'[(12 - sum(
        int(digit) << index
        for index, digit in enumerate(reversed(orcid.replace('-', '')[:-1]), start=1)
    ) % 11) % 11]
    if checksum != orcid[-1]:
        return False, None
    # return sanitized ORCID iD on success
    return True, orcid


@dataclasses.dataclass(frozen=True)
class FederationRef:
    fed_id: int
    component_uuid: str
    referenced_class: typing.Type[typing.Any]

    @dataclasses.dataclass(frozen=True)
    class FederationComponentMock:
        name: str
        address: typing.Optional[str]

        def get_name(self) -> str:
            return self.name

    @property
    def component(self) -> typing.Union[Component, FederationComponentMock]:
        try:
            return get_component_by_uuid(self.component_uuid)
        except errors.ComponentDoesNotExistError:
            return FederationRef.FederationComponentMock(
                name=flask_babel.gettext('Unknown database (%(uuid)s)', uuid=self.component_uuid[:8]),
                address=None
            )


@JinjaFunction()
@dataclasses.dataclass(frozen=True)
class FederationObjectRef(FederationRef):
    referenced_class: typing.Type[Object] = Object


@JinjaFunction()
@dataclasses.dataclass(frozen=True)
class FederationUserRef(FederationRef):
    referenced_class: typing.Type[User] = User


@JinjaFunction()
@dataclasses.dataclass(frozen=True)
class ELNImportObjectRef:
    eln_import_id: int
    eln_object_id: int


@JinjaFunction()
def wrap_object_ref(
        *,
        fed_id: typing.Optional[int] = None,
        component_uuid: typing.Optional[str] = None,
        eln_import_id: typing.Optional[int] = None,
        eln_object_id: typing.Optional[int] = None
) -> typing.Optional[typing.Union[FederationObjectRef, ELNImportObjectRef]]:
    if fed_id is not None and component_uuid is not None:
        return FederationObjectRef(
            fed_id=fed_id,
            component_uuid=component_uuid
        )
    if eln_import_id is not None and eln_object_id is not None:
        return ELNImportObjectRef(
            eln_import_id=eln_import_id,
            eln_object_id=eln_object_id
        )
    return None


@JinjaFunction()
def get_federation_url(obj: typing.Any) -> str:
    component_address: str = obj.component.address
    if not component_address.endswith('/'):
        component_address = component_address + '/'
    if isinstance(obj, FederationRef):
        obj_class = obj.referenced_class
    else:
        obj_class = obj.__class__
    endpoint_and_param_name = {
        User: ('frontend.user_profile', 'user_id'),
        Object: ('frontend.object', 'object_id'),
        Action: ('frontend.action', 'action_id'),
        ActionType: ('frontend.action_type', 'type_id'),
        Instrument: ('frontend.instrument', 'instrument_id'),
        Location: ('frontend.location', 'location_id'),
        LocationType: ('frontend.location_type', 'type_id'),
    }.get(obj_class)
    if endpoint_and_param_name is None:
        return component_address
    else:
        endpoint, param_name = endpoint_and_param_name
        return component_address + relative_url_for(endpoint, **{param_name: obj.fed_id})


@dataclasses.dataclass(frozen=True, kw_only=True)
class LocationFormInformation:
    id: int
    name: str
    name_prefix: str
    id_path: typing.Sequence[int]
    has_subtree: bool
    is_fed: bool
    is_disabled: bool

    @property
    def full_name(self) -> str:
        return self.name_prefix + self.name

    @property
    def id_string(self) -> str:
        return str(self.id)


def get_locations_form_data(
        filter: typing.Callable[[Location], bool]
) -> typing.Tuple[typing.Sequence[LocationFormInformation], typing.Sequence[typing.Tuple[str, str]]]:
    """
    Get location information for a location form field.

    :param filter: a filter for which locations should be valid
    :return: a list of location information and a list of valid choices
    """
    readable_location_ids = {
        location.id
        for location in get_locations_with_user_permissions(current_user.id, Permissions.READ)
    }
    locations_map, locations_tree = get_locations_tree()
    readable_locations_map = {
        location_id: location
        for location_id, location in locations_map.items()
        if location_id in readable_location_ids
    }
    all_choices = [LocationFormInformation(
        id=-1,
        name='—',
        name_prefix='',
        id_path=[],
        has_subtree=False,
        is_fed=False,
        is_disabled=False
    )]
    choices = [
        ('-1', '')
    ]
    unvisited_location_ids_prefixes_and_subtrees = [
        (location_id, '', location, [location_id])
        for location_id, location in locations_tree.items()
    ]
    while unvisited_location_ids_prefixes_and_subtrees:
        location_id, name_prefix, subtree, id_path = unvisited_location_ids_prefixes_and_subtrees.pop(0)
        location = locations_map[location_id]
        # skip hidden locations with a fully hidden subtree
        is_full_subtree_hidden = is_full_location_tree_hidden(readable_locations_map, subtree)
        if not current_user.is_admin and location.is_hidden and is_full_subtree_hidden:
            continue
        has_read_permissions = location_id in readable_location_ids
        # skip unreadable locations, but allow processing their child locations
        # in case any of them are readable
        if has_read_permissions and (not location.is_hidden or current_user.is_admin):
            is_disabled = not filter(location)
            all_choices.append(LocationFormInformation(
                id=location_id,
                name=get_location_name(location, include_id=True, has_read_permissions=has_read_permissions),
                name_prefix=name_prefix,
                id_path=tuple(id_path),
                has_subtree=bool(subtree),
                is_fed=location.fed_id is not None,
                is_disabled=is_disabled
            ))
            if not is_disabled:
                choices.append((str(all_choices[-1].id), all_choices[-1].full_name))
        elif not is_full_subtree_hidden:
            all_choices.append(LocationFormInformation(
                id=location_id,
                name=get_location_name(location, include_id=True, has_read_permissions=has_read_permissions),
                name_prefix=name_prefix,
                id_path=tuple(id_path),
                has_subtree=bool(subtree),
                is_fed=False,
                is_disabled=True,
            ))
        else:
            continue
        name_prefix = f'{name_prefix}{get_location_name(location, has_read_permissions=has_read_permissions)} / '
        for location_id in sorted(subtree, key=lambda location_id: get_location_name(locations_map[location_id], has_read_permissions=location_id in readable_location_ids), reverse=True):
            unvisited_location_ids_prefixes_and_subtrees.insert(0, (location_id, name_prefix, subtree[location_id], id_path + [location_id]))
    return all_choices, choices


@JinjaFunction()
def get_user_or_none(user_id: int, component_id: typing.Optional[int] = None) -> typing.Optional[User]:
    try:
        return get_user(user_id, component_id=component_id)
    except errors.UserDoesNotExistError:
        return None


@dataclasses.dataclass(frozen=True, kw_only=True)
class GroupFormInformation:
    id: int
    name: str
    name_prefix: str
    id_path: typing.Sequence[int]
    has_subtree: bool
    is_disabled: bool

    @property
    def full_name(self) -> str:
        return self.name_prefix + self.name


def get_groups_form_data(
        basic_group_filter: typing.Optional[typing.Callable[[Group], bool]] = None,
        project_group_filter: typing.Optional[typing.Callable[[Project], bool]] = None
) -> typing.Tuple[bool, typing.Sequence[GroupFormInformation]]:
    """
    Get group form information usable with a treepicker, e.g. for permission forms.

    :param basic_group_filter: filter for basic groups to include, or None to
        exclude all basic groups
    :param project_group_filter: filter for project groups to include, or None
        to exclude all project groups
    :return: a tuple containing whether any group is enabled and the group
        form information list
    """
    if basic_group_filter is not None:
        all_basic_groups = get_groups()
        basic_group_names_by_id = {
            group.id: get_translated_text(
                group.name,
                default=flask_babel.gettext('Basic Group') + ' #' + str(group.id)
            )
            for group in all_basic_groups
        }
        valid_basic_group_ids = {
            group.id
            for group in all_basic_groups
            if basic_group_filter(group)
        }
    if project_group_filter is not None:
        all_project_groups = get_projects()
        project_group_names_by_id = {
            group.id: get_translated_text(
                group.name,
                default=flask_babel.gettext('Project Group') + ' #' + str(group.id)
            )
            for group in all_project_groups
        }
        valid_project_group_ids = {
            group.id
            for group in all_project_groups
            if project_group_filter(group)
        }
        child_project_group_ids_by_id = {
            group.id: [
                child_group_id
                for child_group_id in get_child_project_ids(group.id)
                if child_group_id in valid_project_group_ids
            ]
            for group in all_project_groups
        }
    else:
        project_group_names_by_id = {}
        child_project_group_ids_by_id = {}
    all_group_categories = get_group_categories()
    group_category_names_by_id = {
        group_category.id: get_translated_text(
            group_category.name,
            default=flask_babel.gettext('Group Category') + ' #' + str(group_category.id)
        )
        for group_category in all_group_categories
    }
    group_category_tree = get_group_category_tree(
        basic_group_ids=None if basic_group_filter is not None else set(),
        project_group_ids=None if project_group_filter is not None else set()
    )
    all_choices = []

    def _add_project_group_to_all_choices(
            id_path: typing.List[int],
            name_prefix: str,
            group_id: int
    ) -> None:
        if not id_path and get_parent_project_ids(group_id):
            return
        child_project_ids = child_project_group_ids_by_id[group_id]
        project_group_name = project_group_names_by_id[group_id]
        all_choices.append(GroupFormInformation(
            id=group_id,
            name=project_group_name,
            name_prefix=name_prefix,
            id_path=id_path + [group_id],
            has_subtree=bool(child_project_ids),
            is_disabled=group_id not in valid_project_group_ids
        ))
        name_prefix = name_prefix + project_group_name + ' / '
        id_path = id_path + [group_id]
        for child_project_id in sorted(child_project_ids, key=lambda group_id: project_group_names_by_id[group_id]):
            _add_project_group_to_all_choices(
                id_path, name_prefix, child_project_id
            )

    def _fill_all_choices(
            id_path: typing.List[int],
            name_prefix: str,
            group_category_id: typing.Optional[int],
            group_category_tree: GroupCategoryTree
    ) -> None:
        if not group_category_tree['contains_basic_groups'] and not group_category_tree['contains_project_groups']:
            return
        if group_category_id is not None:
            group_category_name = group_category_names_by_id[group_category_id]
            all_choices.append(GroupFormInformation(
                id=-group_category_id,
                name=group_category_name,
                name_prefix=name_prefix,
                id_path=id_path + [-group_category_id],
                has_subtree=True,
                is_disabled=True
            ))
            name_prefix = name_prefix + group_category_name + ' / '
            id_path = id_path + [-group_category_id]
        for group_category_id, group_category_subtree in group_category_tree['child_categories'].items():
            _fill_all_choices(id_path, name_prefix, group_category_id, group_category_subtree)
        if basic_group_filter is not None:
            for group_id in sorted(group_category_tree['basic_group_ids'], key=lambda group_id: basic_group_names_by_id[group_id]):
                all_choices.append(GroupFormInformation(
                    id=group_id,
                    name=basic_group_names_by_id[group_id],
                    name_prefix=name_prefix,
                    id_path=id_path + [group_id],
                    has_subtree=False,
                    is_disabled=group_id not in valid_basic_group_ids
                ))
        if project_group_filter is not None:
            for group_id in sorted(group_category_tree['project_group_ids'], key=lambda group_id: project_group_names_by_id[group_id]):
                _add_project_group_to_all_choices(id_path, name_prefix, group_id)

    _fill_all_choices([], '', None, group_category_tree)

    # filter out subtrees only containing disabled options
    enabled_id_paths = {
        tuple(group_form_information.id_path)
        for group_form_information in all_choices
        if not group_form_information.is_disabled
    }
    all_choices = [
        group_form_information
        for group_form_information in all_choices
        if not group_form_information.is_disabled or (group_form_information.has_subtree and any(
            tuple(group_form_information.id_path) == id_path[:len(group_form_information.id_path)]
            for id_path in enabled_id_paths
        ))
    ]
    any_choice_enabled = any(
        not group_form_information.is_disabled
        for group_form_information in all_choices
    )

    return any_choice_enabled, all_choices


@JinjaFunction()
def get_basic_group_name_prefixes(group_id: int) -> typing.List[str]:
    group_categories = get_basic_group_categories(group_id)
    return sorted([
        ' / '.join(
            get_translated_text(category_name, default=flask_babel.gettext("Unnamed Category"))
            for category_name in get_full_group_category_name(group_category.id)
        ) + ' / '
        for group_category in group_categories
    ])


@JinjaFunction()
def get_project_group_name_prefixes(project_id: int) -> typing.List[str]:
    group_categories = get_project_group_categories(project_id)
    name_prefixes = [
        ' / '.join(
            get_translated_text(category_name, default=flask_babel.gettext("Unnamed Category"))
            for category_name in get_full_group_category_name(group_category.id)
        ) + ' / '
        for group_category in group_categories
    ]
    parent_project_ids = get_parent_project_ids(project_id)
    for parent_project_id in parent_project_ids:
        parent_project = get_project(parent_project_id)
        parent_project_name = get_translated_text(parent_project.name, default=flask_babel.gettext("Unnamed Project Group"))
        parent_project_name_prefixes = get_project_group_name_prefixes(parent_project_id)
        if parent_project_name_prefixes:
            for parent_project_name_prefix in parent_project_name_prefixes:
                name_prefixes.append(parent_project_name_prefix + parent_project_name + ' / ')
        else:
            name_prefixes.append(parent_project_name + ' / ')
    return sorted(name_prefixes)


@JinjaFilter()
def to_timeseries_data(
        data: typing.Dict[str, typing.Any],
        schema: typing.Dict[str, typing.Any]
) -> typing.Dict[str, typing.Any]:
    utc_datetimes: typing.List[datetime] = []
    relative_times: typing.List[typing.Union[int, float]] = []
    if isinstance(data['data'][0][0], str):
        entries = list(sorted(
            (datetime.strptime(entry[0], '%Y-%m-%d %H:%M:%S.%f').replace(tzinfo=timezone.utc), entry[1], entry[2])
            for entry in data['data']
        ))
        utc_datetimes = [
            entry[0]
            for entry in entries
        ]
        magnitudes = [
            entry[1]
            for entry in entries
        ]
    else:
        r_entries = list(sorted((entry[0], entry[1], entry[2]) for entry in data['data']))
        relative_times = [
            entry[0]
            for entry in r_entries
        ]
        magnitudes = [
            entry[1]
            for entry in r_entries
        ]
    display_digits = schema.get('display_digits')
    magnitude_strings = [
        custom_format_number(magnitude, display_digits)
        for magnitude in magnitudes
    ]
    magnitude_average = None
    magnitude_stddev = None
    magnitude_min = None
    magnitude_max = None
    magnitude_count = None
    magnitude_first = None
    magnitude_last = None

    statistics = {'average', 'stddev'}
    if 'statistics' in schema:
        statistics = set(schema['statistics'])
    if 'min' in statistics:
        magnitude_min = np.min(magnitudes)
    if 'max' in statistics:
        magnitude_max = np.max(magnitudes)
    if 'count' in statistics:
        magnitude_count = len(magnitudes)
    if 'first' in statistics:
        magnitude_first = magnitudes[0]
    if 'last' in statistics:
        magnitude_last = magnitudes[-1]

    time_weights = np.zeros(len(magnitudes))
    # determine time-weighted average and standard deviation
    if 'average' in statistics or 'stddev' in statistics:
        if relative_times and len(relative_times) > 1:
            time_weights[0] = (relative_times[1] - relative_times[0]) / 2
            time_weights[-1] = (relative_times[-1] - relative_times[-2]) / 2
            for i in range(1, len(relative_times) - 1):
                rel_previous_time = relative_times[i - 1]
                rel_next_time = relative_times[i + 1]
                time_weights[i] = (rel_next_time - rel_previous_time) / 2
            magnitude_average = np.average(magnitudes, weights=time_weights)
            if 'stddev' in statistics:
                magnitude_stddev = np.sqrt(np.average(np.square(np.array(magnitudes) - magnitude_average), weights=time_weights))
            if 'average' not in statistics:
                magnitude_average = None
        elif len(utc_datetimes) > 1:
            time_weights[0] = (utc_datetimes[1] - utc_datetimes[0]).total_seconds() / 2
            time_weights[-1] = (utc_datetimes[-1] - utc_datetimes[-2]).total_seconds() / 2
            for i in range(1, len(utc_datetimes) - 1):
                previous_time = utc_datetimes[i - 1]
                next_time = utc_datetimes[i + 1]
                time_weights[i] = (next_time - previous_time).total_seconds() / 2
            magnitude_average = np.average(magnitudes, weights=time_weights)
            if 'stddev' in statistics:
                magnitude_stddev = np.sqrt(np.average(np.square(np.array(magnitudes) - magnitude_average), weights=time_weights))
            if 'average' not in statistics:
                magnitude_average = None
        else:
            if 'average' in statistics:
                magnitude_average = magnitudes[0]
    same_day = None
    if utc_datetimes:
        # convert datetimes to local timezone
        local_timezone = pytz.timezone(current_user.timezone or 'UTC')
        local_datetimes = [
            utc_datetime.astimezone(local_timezone)
            for utc_datetime in utc_datetimes
        ]
        time_strings = [
            local_datetime.strftime('%Y-%m-%d %H:%M:%S.%f')
            for local_datetime in local_datetimes
        ]
        same_day = utc_datetimes[0].date() == utc_datetimes[-1].date()
    else:
        time_strings = [
            custom_format_number(time, display_digits) + ' s'
            for time in relative_times
        ]
    return {
        'title': get_translated_text(schema.get('title')),
        'units': ' ' + prettify_units(data['units']) if data['dimensionality'] != 'dimensionless' else '',
        'relative_times': relative_times,
        'utc_datetimes': utc_datetimes,
        'times_strings': time_strings,
        'magnitudes': magnitudes,
        'magnitude_strings': magnitude_strings,
        'magnitude_average': magnitude_average,
        'magnitude_stddev': magnitude_stddev,
        'magnitude_min': magnitude_min,
        'magnitude_max': magnitude_max,
        'magnitude_count': magnitude_count,
        'magnitude_first': magnitude_first,
        'magnitude_last': magnitude_last,
        'same_day': same_day
    }


@JinjaFilter()
def to_timeseries_csv(
        rows: typing.List[typing.List[typing.Union[str, float]]]
) -> str:
    rows = copy.deepcopy(rows)
    user_timezone = pytz.timezone(current_user.timezone or 'UTC')
    if isinstance(rows[0][0], str):
        if user_timezone != pytz.utc:
            # convert datetimes from UTC if necessary
            for row in rows:
                try:
                    utc_datetime = datetime.strptime(typing.cast(str, row[0]), '%Y-%m-%d %H:%M:%S.%f').replace(tzinfo=timezone.utc)
                except ValueError:
                    continue
                local_datetime = utc_datetime.astimezone(user_timezone)
                row[0] = local_datetime.strftime('%Y-%m-%d %H:%M:%S.%f')
    csv_file = io.StringIO()
    writer = csv.writer(csv_file, quoting=csv.QUOTE_NONNUMERIC)
    writer.writerows(rows)
    return csv_file.getvalue()


@JinjaFilter()
def encode_choices(condition_list: list[dict[str, typing.Any]]) -> list[dict[str, typing.Any]]:
    for condition in condition_list:
        if condition['type'] == 'choice_equals':
            condition['encoded_choice'] = base64encode(condition['choice'])
        if condition['type'] == 'not':
            condition['condition'] = encode_choices([condition['condition']])[0]
        if condition['type'] in ('any', 'all'):
            condition['conditions'] = encode_choices(condition['conditions'])
    return condition_list


@JinjaFilter()
def stringify(json_object: list[dict[str, typing.Any]]) -> str:
    return json.dumps(json_object, separators=(',', ':'))


@JinjaFilter()
def unify_url(url: str) -> str:
    if not url.endswith('/'):
        url += '/'
    return url


@JinjaFunction()
def to_diff_table(
        lines_before: str,
        lines_after: str,
        label_before: str = '',
        label_after: str = ''
) -> str:
    return difflib.HtmlDiff(wrapcolumn=55).make_table(fromlines=lines_before.splitlines(), tolines=lines_after.splitlines(), fromdesc=label_before, todesc=label_after)


@JinjaFunction('get_component_by_uuid')
def safe_get_component_by_uuid(component_uuid: str) -> typing.Optional[Component]:
    try:
        return get_component_by_uuid(component_uuid=component_uuid)
    except errors.ComponentDoesNotExistError:
        return None


@JinjaFunction()
def get_property_names_in_order(
        schema: typing.Dict[str, typing.Any],
        previous_schema: typing.Optional[typing.Dict[str, typing.Any]] = None
) -> typing.Sequence[str]:
    property_names = [
        property_name
        for property_name in schema.get('propertyOrder', [])
        if property_name in schema.get('properties', [])
    ] + [
        property_name
        for property_name in schema.get('properties', [])
        if property_name not in schema.get('propertyOrder', [])
    ]
    if previous_schema:
        for property_name in get_property_names_in_order(previous_schema):
            if property_name not in property_names:
                property_names.append(property_name)
    return property_names


@JinjaFunction()
def id_prefix_for_property_path(
        property_path: typing.Tuple[typing.Union[str, int], ...],
        id_prefix_root: str
) -> str:
    return '__'.join([
        str(path_element)
        for path_element in [id_prefix_root] + list(property_path)
    ]) + '_'


@JinjaFunction()
def is_deep_diff_possible(
        diff: DataDiff,
        schema: typing.Dict[str, typing.Any],
        previous_schema: typing.Dict[str, typing.Any]
) -> bool:
    if '_before' in diff or '_after' in diff:
        return False
    if schema.get('type') == previous_schema.get('type') == 'array':
        if schema.get('style') != previous_schema.get('style'):
            return False
        if schema.get('style') == 'table' and schema.get('items', {}).get('type') != previous_schema.get('items', {}).get('type'):
            return False
    return True


@JinjaFunction()
def set_template_value(
        name: str,
        value: typing.Any
) -> None:
    if type(value) is jinja2.runtime.Undefined:
        # do not store undefined values, which will result in undefined as result in JavaScript
        return
    if not hasattr(flask.g, 'template_values'):
        flask.g.template_values = {}
    if flask.g.template_values is None:
        raise errors.TemplateValueError()
    if name in flask.g.template_values:
        raise errors.TemplateValueError()
    flask.g.template_values[name] = value


@JinjaFunction()
def get_template_values() -> typing.Any:
    if not hasattr(flask.g, 'template_values'):
        flask.g.template_values = {}
    template_values = flask.g.template_values
    flask.g.template_values = None
    return template_values


@JinjaFunction()
def current_utc_datetime() -> datetime:
    return datetime.now(tz=timezone.utc).replace(tzinfo=None)


@JinjaFunction()
@cache_per_request()
def get_federated_identity(user: User | int) -> tuple[User, typing.Optional[User]]:
    if isinstance(user, int):
        user = get_user(user)
    if user is None:
        return user, None
    federated_user = get_user_by_federated_user(federated_user_id=user.id)
    if federated_user is not None:
        return federated_user, user
    return user, None


def build_modified_url(
        endpoint: str,
        blocked_parameters: typing.Sequence[str] = (),
        **query_parameters: typing.Any
) -> str:
    for param in flask.request.args:
        if param not in query_parameters:
            query_parameters[param] = flask.request.args.getlist(param)
    for param in blocked_parameters:
        if param in query_parameters:
            del query_parameters[param]
    return flask.url_for(
        endpoint,
        **query_parameters
    )


def parse_filter_id_params(
        params: werkzeug.datastructures.MultiDict[str, str],
        param_aliases: typing.List[str],
        valid_ids: typing.List[int],
        id_map: typing.Dict[str, int],
        multi_params_error: str,
        parse_error: str,
        invalid_id_error: str
) -> typing.Tuple[bool, typing.Optional[typing.List[int]]]:
    num_used_param_aliases = sum(param_alias in params for param_alias in param_aliases)
    if num_used_param_aliases == 0:
        return True, None
    if num_used_param_aliases > 1:
        flask.flash(multi_params_error, 'error')
        return False, None
    try:
        filter_ids = set()
        for param_alias in param_aliases:
            for ids_str in params.getlist(param_alias):
                for id_str in ids_str.split(','):
                    id_str = id_str.strip()
                    if not id_str:
                        continue
                    if id_str in id_map:
                        filter_ids.add(id_map[id_str])
                    else:
                        filter_ids.add(int(id_str))
    except ValueError:
        flask.flash(parse_error, 'error')
        return False, None
    if any(id not in valid_ids for id in filter_ids):
        flask.flash(invalid_id_error, 'error')
        return False, None
    if not filter_ids:
        return True, None
    return True, list(filter_ids)


@JinjaFilter()
def timeline_array_to_plotly_chart(
        timeline_array: typing.List[typing.Dict[str, typing.Any]]
) -> typing.Dict[str, typing.Any]:
    utc_datetime_strings = []
    for item in timeline_array:
        utc_datetime_strings.append(item['datetime']['utc_datetime'])
    markers = []
    max_concurrent_count = 1
    for index, item in enumerate(timeline_array):
        utc_datetime_string = item['datetime']['utc_datetime']
        concurrent_item_count = utc_datetime_strings.count(utc_datetime_string)
        utc_datetime = datetime.strptime(utc_datetime_string, '%Y-%m-%d %H:%M:%S').replace(tzinfo=timezone.utc)
        local_datetime = utc_datetime.astimezone(pytz.timezone(current_user.timezone or 'UTC'))
        local_datetime_string = local_datetime.strftime('%Y-%m-%d %H:%M:%S')
        if isinstance(item.get('label'), dict) and item['label'].get('_type') == 'text' and not item['label'].get('multiline') and not item['label'].get('markdown') and get_translated_text(item['label'].get('text', '')).strip():
            label = typing.cast(str, markupsafe.escape(get_translated_text(item['label'].get('text', '')).strip()))
        else:
            label = format_datetime(local_datetime, format='medium')
        if concurrent_item_count > 1:
            max_concurrent_count = max(max_concurrent_count, concurrent_item_count)
            concurrent_item_index = utc_datetime_strings[:index].count(utc_datetime_string)
            markers.append((local_datetime_string, concurrent_item_index, label))
        else:
            markers.append((local_datetime_string, 0, label))
    plotly_chart = {
        "data": [
            {
                "x": [
                    marker[0]
                    for marker in markers
                ],
                "y": [
                    str(marker[1])
                    for marker in markers
                ],
                "text": [
                    marker[2]
                    for marker in markers
                ],
                "mode": "markers+text",
                "hoverinfo": "none",
                "textposition": "top center",
                "type": "scatter"
            }
        ],
        "layout": {
            "font": {
                "family": "Arial, sans-serif",
                "size": 12,
                "color": "#000"
            },
            "showlegend": False,
            "autosize": False,
            "width": 800,
            "height": 175 + 25 * max_concurrent_count,
            "xaxis": {
                "showgrid": False
            },
            "yaxis": {
                "showgrid": False,
                "zeroline": True,
                "showticklabels": False,
                "range": [
                    -0.5,
                    max_concurrent_count
                ]
            },
            "margin": {
                "l": 70,
                "r": 40,
                "b": 60,
                "t": 60,
                "pad": 2,
                "autoexpand": True
            },
            "dragmode": "pan",
            "paper_bgcolor": "#fff",
            "plot_bgcolor": "#fff",
            "hovermode": "closest",
            "separators": ".,"
        }
    }
    return plotly_chart


@JinjaFunction()
def get_action_type_ids_with_usable_actions() -> typing.Set[int]:
    action_type_ids_with_usable_actions = set()
    if current_user.is_authenticated and not current_user.is_readonly:
        actions = get_actions_with_permissions(current_user.id, permissions=Permissions.READ)
        for action in actions:
            if action.type is None:
                continue
            if action.admin_only and not current_user.is_admin:
                continue
            if action.disable_create_objects:
                continue
            if action.type.disable_create_objects:
                continue
            action_type_ids_with_usable_actions.add(action.type.id)
    return action_type_ids_with_usable_actions


@JinjaFunction()
def get_info_pages() -> typing.Sequence[InfoPage]:
    if not current_user.is_authenticated or current_user.is_readonly:
        return []
    return get_info_pages_for_endpoint(
        endpoint=flask.request.endpoint,
        user_id=current_user.id,
        exclude_disabled=True
    )



# coding: utf-8
"""
Add binary data column to files table.
"""

import flask_sqlalchemy

from .utils import table_has_column


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if table_has_column('files', 'binary_data'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE files
        ADD binary_data BYTEA NULL
    """))
    return True



import typing

import pytest

import sampledb.config

@pytest.mark.parametrize(
    ['default_notification_modes', 'should_be_valid'],
    [
        (None, True),
        (
            {
                'DEFAULT': 'WEBAPP'
            },
            True
        ),
        (
            {
                'DEFAULT': 'EMAIL'
            },
            True
        ),
        (
            {
                'DEFAULT': 'IGNORE'
            },
            True
        ),
        (
            {
                'DEFAULT': 'OTHER'
            },
            False
        ),
        (
            {
                'ANNOUNCEMENT': 'WEBAPP'
            },
            True
        ),
        (
            {
                'announcement': 'WEBAPP'
            },
            False
        )
    ]
)
def test_is_default_notification_modes_valid(default_notification_modes: typing.Any, should_be_valid: bool):
    sampledb.config.DEFAULT_NOTIFICATION_MODES = default_notification_modes
    assert sampledb.config.is_default_notification_modes_valid() == should_be_valid



# coding: utf-8
"""
Add INSTRUMENT_LOG_ENTRY_CREATED enum value to NotificationType enum.
"""

import flask_sqlalchemy

from .utils import enum_value_migration


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    return enum_value_migration('notificationtype', 'INSTRUMENT_LOG_ENTRY_CREATED')



import typing

import itsdangerous

MAX_AGE = 60 * 60 * 24 * 2


def generate_token(
        data: typing.Any,
        salt: str,
        secret_key: str
) -> str:
    serializer = itsdangerous.URLSafeTimedSerializer(secret_key)
    return serializer.dumps(data, salt=salt)


def verify_token(
        token: str,
        salt: str,
        secret_key: str,
        expiration: typing.Optional[int] = MAX_AGE
) -> typing.Optional[typing.Any]:
    serializer = itsdangerous.URLSafeTimedSerializer(secret_key)
    try:
        return serializer.loads(
            token,
            salt=salt,
            max_age=expiration
        )
    except itsdangerous.BadData:
        return None



# coding: utf-8
"""

"""

import typing

from sqlalchemy.orm import relationship, Mapped, Query

from .. import db
from .utils import Model

if typing.TYPE_CHECKING:
    from .components import Component
    from .locations import Location
    from .instrument_translation import InstrumentTranslation
    from .users import User
    from .topics import Topic

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'

instrument_user_association_table = db.Table(
    'association',
    db.metadata,
    db.Column('instrument_id', db.Integer, db.ForeignKey('instruments.id')),
    db.Column('user_id', db.Integer, db.ForeignKey('users.id'))
)

topic_instrument_association_table = db.Table(
    'instrument_topics',
    db.metadata,
    db.Column('instrument_id', db.Integer, db.ForeignKey('instruments.id')),
    db.Column('topic_id', db.Integer, db.ForeignKey('topics.id'))
)


class Instrument(Model):
    __tablename__ = 'instruments'
    __table_args__ = (
        db.UniqueConstraint('fed_id', 'component_id', name='instruments_fed_id_component_id_key'),
    )

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    responsible_users: Mapped[typing.List['User']] = relationship("User", secondary=instrument_user_association_table, order_by="User.name", lazy='selectin')
    users_can_create_log_entries: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False)
    users_can_view_log_entries: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False)
    create_log_entry_default: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False)
    is_hidden: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False)
    notes_is_markdown: Mapped[typing.Optional[bool]] = db.Column(db.Boolean, nullable=True, default=False)
    description_is_markdown: Mapped[typing.Optional[bool]] = db.Column(db.Boolean, nullable=True, default=False)
    short_description_is_markdown: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False)
    fed_id: Mapped[typing.Optional[int]] = db.Column(db.Integer, nullable=True)
    component_id: Mapped[typing.Optional[int]] = db.Column(db.Integer, db.ForeignKey('components.id'), nullable=True)
    component: Mapped[typing.Optional['Component']] = relationship('Component')
    translations: Mapped[typing.List['InstrumentTranslation']] = relationship('InstrumentTranslation', lazy='selectin')
    location_id: Mapped[typing.Optional[int]] = db.Column(db.Integer, db.ForeignKey('locations.id'), nullable=True)
    location: Mapped[typing.Optional['Location']] = relationship('Location')
    object_id: Mapped[typing.Optional[int]] = db.Column(db.Integer, db.ForeignKey('objects_current.object_id'), nullable=True)
    show_linked_object_data: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=True, server_default=db.true())
    topics: Mapped[typing.List['Topic']] = relationship('Topic', secondary=topic_instrument_association_table, back_populates='instruments')

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["Instrument"]]

    def __init__(
            self,
            description_is_markdown: bool = False,
            short_description_is_markdown: bool = False,
            notes_is_markdown: bool = False,
            users_can_create_log_entries: bool = False,
            users_can_view_log_entries: bool = False,
            create_log_entry_default: bool = False,
            is_hidden: bool = False,
            fed_id: typing.Optional[int] = None,
            component_id: typing.Optional[int] = None,
            show_linked_object_data: bool = True
    ) -> None:
        super().__init__(
            description_is_markdown=description_is_markdown,
            short_description_is_markdown=short_description_is_markdown,
            notes_is_markdown=notes_is_markdown,
            users_can_create_log_entries=users_can_create_log_entries,
            users_can_view_log_entries=users_can_view_log_entries,
            create_log_entry_default=create_log_entry_default,
            is_hidden=is_hidden,
            fed_id=fed_id,
            component_id=component_id,
            show_linked_object_data=show_linked_object_data,
        )

    def __eq__(self, other: typing.Any) -> bool:
        if isinstance(other, Instrument):
            return bool(
                self.id == other.id and
                self.description_is_markdown == other.description_is_markdown and
                self.short_description_is_markdown == other.short_description_is_markdown and
                self.notes_is_markdown == other.notes_is_markdown and
                self.users_can_create_log_entries == other.users_can_create_log_entries and
                self.users_can_view_log_entries == other.users_can_view_log_entries and
                self.create_log_entry_default == other.create_log_entry_default and
                self.is_hidden == other.is_hidden and
                self.responsible_users == other.responsible_users and
                self.fed_id == other.fed_id and
                self.component_id == other.component_id
            )
        return NotImplemented

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id})>'

    @property
    def name(self) -> typing.Dict[str, str]:
        return {
            translation.language.lang_code: translation.name
            for translation in self.translations
            if translation.name
        }

    @property
    def notes(self) -> typing.Dict[str, str]:
        return {
            translation.language.lang_code: translation.notes
            for translation in self.translations
            if translation.notes
        }

    @property
    def description(self) -> typing.Dict[str, str]:
        return {
            translation.language.lang_code: translation.description
            for translation in self.translations
            if translation.description
        }

    @property
    def short_description(self) -> typing.Dict[str, str]:
        return {
            translation.language.lang_code: translation.short_description
            for translation in self.translations
            if translation.short_description
        }



import typing

import flask

from .components import _get_or_create_component_id
from .utils import _get_id, _get_uuid, _get_bool, _get_str, _get_dict
from ..languages import get_languages, get_language, get_language_by_lang_code
from ..action_types import ActionType, get_action_type, create_action_type, update_action_type
from ..action_type_translations import set_action_type_translation, get_action_type_translations_for_action_type
from ..components import Component
from .. import errors, fed_logs


class ActionTypeRef(typing.TypedDict):
    action_type_id: int
    component_uuid: str


class ActionTypeTranslationData(typing.TypedDict):
    language_id: int
    name: str
    description: str
    object_name: str
    object_name_plural: str
    view_text: str
    perform_text: str


class ActionTypeData(typing.TypedDict):
    fed_id: int
    component_uuid: str
    admin_only: bool
    enable_labels: bool
    enable_files: bool
    enable_locations: bool
    enable_publications: bool
    enable_comments: bool
    enable_activity_log: bool
    enable_related_objects: bool
    enable_project_link: bool
    enable_instrument_link: bool
    disable_create_objects: bool
    is_template: bool
    translations: typing.List[ActionTypeTranslationData]


class SharedActionTypeTranslationData(typing.TypedDict):
    name: str
    description: str
    object_name: str
    object_name_plural: str
    view_text: str
    perform_text: str


class SharedActionTypeData(typing.TypedDict):
    action_type_id: int
    component_uuid: str
    admin_only: bool
    enable_labels: bool
    enable_files: bool
    enable_locations: bool
    enable_publications: bool
    enable_comments: bool
    enable_activity_log: bool
    enable_related_objects: bool
    enable_project_link: bool
    enable_instrument_link: bool
    disable_create_objects: bool
    is_template: bool
    translations: typing.Optional[typing.Dict[str, SharedActionTypeTranslationData]]


def import_action_type(
        action_type_data: ActionTypeData,
        component: Component
) -> ActionType:
    component_id = _get_or_create_component_id(action_type_data['component_uuid'])
    # component_id will only be None if this would import a local action type
    assert component_id is not None

    try:
        action_type = get_action_type(action_type_data['fed_id'], component_id)

        ignored_keys = {
            'fed_id',
            'component_uuid',
            'translations'
        }
        if any(
                value != getattr(action_type, key)
                for key, value in action_type_data.items()
                if key not in ignored_keys
        ):
            update_action_type(
                action_type_id=action_type.id,
                admin_only=action_type_data['admin_only'],
                show_on_frontpage=False,
                show_in_navbar=False,
                show_in_object_filters=True,
                enable_labels=action_type_data['enable_labels'],
                enable_files=action_type_data['enable_files'],
                enable_locations=action_type_data['enable_locations'],
                enable_publications=action_type_data['enable_publications'],
                enable_comments=action_type_data['enable_comments'],
                enable_activity_log=action_type_data['enable_activity_log'],
                enable_related_objects=action_type_data['enable_related_objects'],
                enable_project_link=action_type_data['enable_project_link'],
                enable_instrument_link=action_type_data['enable_instrument_link'],
                disable_create_objects=action_type_data['disable_create_objects'],
                is_template=action_type_data['is_template'],
                scicat_export_type=None
            )
            fed_logs.update_action_type(action_type.id, component.id)
    except errors.ActionTypeDoesNotExistError:
        action_type = create_action_type(
            fed_id=action_type_data['fed_id'],
            component_id=component_id,
            admin_only=action_type_data['admin_only'],
            show_on_frontpage=False,
            show_in_navbar=False,
            show_in_object_filters=True,
            enable_labels=action_type_data['enable_labels'],
            enable_files=action_type_data['enable_files'],
            enable_locations=action_type_data['enable_locations'],
            enable_publications=action_type_data['enable_publications'],
            enable_comments=action_type_data['enable_comments'],
            enable_activity_log=action_type_data['enable_activity_log'],
            enable_related_objects=action_type_data['enable_related_objects'],
            enable_project_link=action_type_data['enable_project_link'],
            enable_instrument_link=action_type_data['enable_instrument_link'],
            disable_create_objects=action_type_data['disable_create_objects'],
            is_template=action_type_data['is_template'],
            scicat_export_type=None
        )
        fed_logs.import_action_type(action_type.id, component.id)

    for action_type_translation in action_type_data['translations']:
        set_action_type_translation(
            action_type_id=action_type.id,
            language_id=action_type_translation['language_id'],
            name=action_type_translation['name'],
            description=action_type_translation['description'],
            object_name=action_type_translation['object_name'],
            object_name_plural=action_type_translation['object_name_plural'],
            view_text=action_type_translation['view_text'],
            perform_text=action_type_translation['perform_text']
        )
    return action_type


def parse_action_type(
        action_type_data: typing.Dict[str, typing.Any]
) -> ActionTypeData:
    fed_id = _get_id(action_type_data.get('action_type_id'), special_values=[ActionType.SAMPLE_CREATION, ActionType.MEASUREMENT, ActionType.SIMULATION, ActionType.TEMPLATE])
    uuid = _get_uuid(action_type_data.get('component_uuid'))
    if uuid == flask.current_app.config['FEDERATION_UUID']:
        # do not accept updates for own data
        raise errors.InvalidDataExportError(f'Invalid update for local action type {fed_id}')
    result = ActionTypeData(
        fed_id=fed_id,
        component_uuid=uuid,
        admin_only=_get_bool(action_type_data.get('admin_only'), default=True),
        enable_labels=_get_bool(action_type_data.get('enable_labels'), default=False),
        enable_files=_get_bool(action_type_data.get('enable_files'), default=False),
        enable_locations=_get_bool(action_type_data.get('enable_locations'), default=False),
        enable_publications=_get_bool(action_type_data.get('enable_publications'), default=False),
        enable_comments=_get_bool(action_type_data.get('enable_comments'), default=False),
        enable_activity_log=_get_bool(action_type_data.get('enable_activity_log'), default=False),
        enable_related_objects=_get_bool(action_type_data.get('enable_related_objects'), default=False),
        enable_project_link=_get_bool(action_type_data.get('enable_project_link'), default=False),
        enable_instrument_link=_get_bool(action_type_data.get('enable_instrument_link'), default=False),
        disable_create_objects=_get_bool(action_type_data.get('disable_create_objects'), default=False),
        is_template=_get_bool(action_type_data.get('is_template'), default=False),
        translations=[]
    )

    allowed_language_ids = [language.id for language in get_languages(only_enabled_for_input=False)]

    translation_data = _get_dict(action_type_data.get('translations'))
    if translation_data is not None:
        for lang_code, translation in translation_data.items():
            try:
                language = get_language_by_lang_code(lang_code.lower())
            except errors.LanguageDoesNotExistError:
                continue
            language_id = language.id
            if language_id not in allowed_language_ids:
                continue

            result['translations'].append(ActionTypeTranslationData(
                language_id=language_id,
                name=_get_str(translation.get('name'), default=''),
                description=_get_str(translation.get('description'), default=''),
                object_name=_get_str(translation.get('object_name'), default=''),
                object_name_plural=_get_str(translation.get('object_name_plural'), default=''),
                view_text=_get_str(translation.get('view_text'), default=''),
                perform_text=_get_str(translation.get('perform_text'), default='')
            ))
    return result


def parse_import_action_type(
        action_type_data: typing.Dict[str, typing.Any],
        component: Component
) -> ActionType:
    return import_action_type(parse_action_type(action_type_data), component)


def _parse_action_type_ref(
        action_type_data: typing.Union[ActionTypeRef, typing.Dict[str, typing.Any]]
) -> ActionTypeRef:
    action_type_id = _get_id(
        action_type_data.get('action_type_id'),
        special_values=[
            ActionType.SAMPLE_CREATION,
            ActionType.MEASUREMENT,
            ActionType.SIMULATION,
            ActionType.TEMPLATE
        ]
    )
    component_uuid = _get_uuid(action_type_data.get('component_uuid'))
    return ActionTypeRef(
        action_type_id=action_type_id,
        component_uuid=component_uuid
    )


def _get_or_create_action_type_id(
        action_type_data: typing.Optional[ActionTypeRef]
) -> typing.Optional[int]:
    if action_type_data is None:
        return None
    component_id = _get_or_create_component_id(action_type_data['component_uuid'])
    try:
        action_type = get_action_type(action_type_data['action_type_id'], component_id)
    except errors.ActionTypeDoesNotExistError:
        assert component_id is not None
        action_type = create_action_type(
            admin_only=True,
            show_on_frontpage=False,
            show_in_navbar=False,
            show_in_object_filters=True,
            enable_labels=False,
            enable_files=False,
            enable_locations=False,
            enable_publications=False,
            enable_comments=False,
            enable_activity_log=False,
            enable_related_objects=False,
            enable_project_link=False,
            enable_instrument_link=False,
            disable_create_objects=False,
            is_template=False,
            fed_id=action_type_data['action_type_id'],
            component_id=component_id,
            scicat_export_type=None
        )
        fed_logs.create_ref_action_type(action_type.id, component_id)
    return action_type.id


def shared_action_type_preprocessor(
        action_type_id: int,
        _component: Component,
        _refs: typing.List[typing.Tuple[str, int]],
        _markdown_images: typing.Dict[str, str]
) -> typing.Optional[SharedActionTypeData]:
    action_type = get_action_type(action_type_id)
    if action_type.component_id is not None:
        return None
    translations_data = {}
    try:
        translations = get_action_type_translations_for_action_type(action_type.id)
        for translation in translations:
            lang_code = get_language(translation.language_id).lang_code
            translations_data[lang_code] = SharedActionTypeTranslationData(
                name=translation.name,
                description=translation.description,
                object_name=translation.object_name,
                object_name_plural=translation.object_name_plural,
                view_text=translation.view_text,
                perform_text=translation.perform_text
            )
    except errors.ActionTypeTranslationDoesNotExistError:
        translations_data = {}
    return SharedActionTypeData(
        action_type_id=action_type.id if action_type.component_id is None else action_type.fed_id,
        component_uuid=flask.current_app.config['FEDERATION_UUID'] if action_type.component is None else action_type.component.uuid,
        admin_only=action_type.admin_only,
        enable_labels=action_type.enable_labels,
        enable_files=action_type.enable_files,
        enable_locations=action_type.enable_locations,
        enable_publications=action_type.enable_publications,
        enable_comments=action_type.enable_comments,
        enable_activity_log=action_type.enable_activity_log,
        enable_related_objects=action_type.enable_related_objects,
        enable_project_link=action_type.enable_project_link,
        enable_instrument_link=action_type.enable_instrument_link,
        disable_create_objects=action_type.disable_create_objects,
        is_template=action_type.is_template,
        translations=translations_data if translations_data else None
    )



import typing

import flask

from . import languages

SUPPORTED_LOCALES: typing.Dict[str, typing.Dict[str, str]] = {
    'de': {"english_name": "German", "native_name": "Deutsch"},
    'en': {"english_name": "English", "native_name": "English"}
}

DEFAULT_LOCALE = 'en'


def guess_request_locale() -> str:
    """
    Guess the best locale for a flask request.

    If no match is found, the default locale (English) is returned.

    :return: the locale language code
    """
    allowed_language_codes = get_allowed_language_codes()
    try:
        best_match_locale: typing.Optional[str] = flask.request.accept_languages.best_match(allowed_language_codes)
    except Exception:
        return DEFAULT_LOCALE
    if best_match_locale is not None and best_match_locale in SUPPORTED_LOCALES:
        return best_match_locale
    else:
        return DEFAULT_LOCALE


def get_allowed_language_codes() -> typing.List[str]:
    supported_language_codes = list(SUPPORTED_LOCALES.keys())
    allowed_language_codes = languages.get_language_codes(only_enabled_for_user_interface=True)
    return [
        lang_code
        for lang_code in supported_language_codes
        if lang_code in allowed_language_codes
    ]



# coding: utf-8
"""

"""

import enum
import typing
import datetime

from sqlalchemy.orm import Query, Mapped

from .. import db
from .utils import Model


class WebhookType(enum.Enum):
    OBJECT_LOG = 1


class Webhook(Model):
    __tablename__ = "webhooks"

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    user_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False)
    type: Mapped[WebhookType] = db.Column(db.Enum(WebhookType), nullable=False)
    target_url: Mapped[str] = db.Column(db.String, nullable=False)
    name: Mapped[typing.Optional[str]] = db.Column(db.String, default=None, nullable=True)
    secret: Mapped[str] = db.Column(db.String, nullable=False)
    last_contact: Mapped[typing.Optional[datetime.datetime]] = db.Column(db.DateTime, default=None, nullable=True)
    __table_args__ = (
        db.UniqueConstraint('user_id', 'type', 'target_url', name='_user_id_type_target_url_uc'),
    )

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["Webhook"]]

    def __init__(
        self,
        user_id: int,
        type: WebhookType,
        target_url: str,
        secret: str,
        name: typing.Optional[str],
        last_contact: typing.Optional[datetime.datetime] = None
    ) -> None:
        super().__init__(
            user_id=user_id,
            type=type,
            target_url=target_url,
            name=name,
            secret=secret,
            last_contact=last_contact
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id})>'



# coding: utf-8
"""
RESTful API for SampleDB
"""

import datetime
import typing

import flask

from ..utils import Resource, ResponseData
from ...logic import errors
from ...logic.components import Component, get_components, get_component_infos
from ...logic.shares import get_shares_for_component, get_share, ObjectShare, set_object_share_import_status, parse_object_share_import_status
from ...logic.federation.action_types import shared_action_type_preprocessor
from ...logic.federation.actions import shared_action_preprocessor
from ...logic.federation.instruments import shared_instrument_preprocessor
from ...logic.federation.location_types import shared_location_type_preprocessor
from ...logic.federation.locations import shared_location_preprocessor
from ...logic.federation.login import get_idp_metadata, get_sp_metadata, check_component_locally_ready
from ...logic.federation.objects import shared_object_preprocessor
from ...logic.federation.update import import_updates, PROTOCOL_VERSION_MAJOR, PROTOCOL_VERSION_MINOR
from ...logic.federation.users import shared_user_preprocessor
from ...api.federation.authentication import http_token_auth
from ...logic.files import get_file
from ...logic.users import get_user_aliases_for_component, get_users, get_user_email_hashes, User
from ...models import UserType

preprocessors = {
    'actions': shared_action_preprocessor,
    'users': shared_user_preprocessor,
    'instruments': shared_instrument_preprocessor,
    'locations': shared_location_preprocessor,
    'location_types': shared_location_type_preprocessor,
    'action_types': shared_action_type_preprocessor
}


def _get_header(component: Component) -> typing.Dict[str, typing.Any]:
    return {
        'db_uuid': flask.current_app.config['FEDERATION_UUID'],
        'target_uuid': component.uuid,
        'protocol_version': {
            'major': PROTOCOL_VERSION_MAJOR,
            'minor': PROTOCOL_VERSION_MINOR
        },
        'sync_timestamp': datetime.datetime.now(datetime.timezone.utc).strftime('%Y-%m-%d %H:%M:%S.%f'),
    }


def _get_last_sync(args: typing.Dict[str, typing.Any]) -> typing.Optional[datetime.datetime]:
    if 'last_sync_timestamp' not in args.keys():
        return None
    try:
        return datetime.datetime.strptime(args['last_sync_timestamp'], '%Y-%m-%d %H:%M:%S.%f').replace(tzinfo=datetime.timezone.utc)
    except ValueError:
        return None


def share_to_json(share: ObjectShare) -> typing.Dict[str, typing.Any]:
    return {
        'object_id': share.object_id,
        'policy': share.policy,
        'utc_datetime': share.utc_datetime.strftime('%Y-%m-%d %H:%M:%S.%f') if share.utc_datetime else None
    }


class UpdateHook(Resource):
    @http_token_auth.login_required
    def post(self) -> ResponseData:
        try:
            import_updates(flask.g.component)
        except errors.UnauthorizedRequestError:
            pass
        except errors.NoAuthenticationMethodError:
            pass
        except errors.InvalidDataExportError:
            pass
        except errors.MissingComponentAddressError:
            pass
        except errors.ComponentNotConfiguredForFederationError:
            pass
        except errors.RequestServerError:
            pass
        except errors.RequestError:
            pass
        except ConnectionError:
            pass
        return None


class ImportStatus(Resource):
    @http_token_auth.login_required
    def put(self, object_id: int) -> ResponseData:
        component_id = flask.g.component.id
        try:
            get_share(object_id, component_id)
        except errors.ObjectDoesNotExistError:
            return {
                "message": f"object {object_id} does not exist"
            }, 404
        except errors.ShareDoesNotExistError:
            return {
                "message": f"object {object_id} is not shared"
            }, 403
        import_status = flask.request.json
        if not import_status:
            return {
                "message": "missing import status"
            }, 400
        parsed_import_status = parse_object_share_import_status(import_status)
        if parsed_import_status is None:
            return {
                "message": "invalid import status"
            }, 400
        set_object_share_import_status(
            object_id=object_id,
            component_id=component_id,
            import_status=dict(parsed_import_status)
        )
        return '', 204


class Objects(Resource):
    @http_token_auth.login_required
    def get(self) -> ResponseData:
        component = flask.g.component
        shares = get_shares_for_component(component.id)
        refs: typing.List[typing.Tuple[str, int]] = []
        markdown_images: typing.Dict[str, str] = {}
        ref_ids: typing.Dict[str, typing.List[int]] = {
            'actions': [],
            'users': [],
            'instruments': [],
            'locations': [],
            'location_types': [],
            'action_types': []
        }

        result_lists: typing.Dict[str, typing.List[typing.Any]] = {
            'actions': [],
            'users': [],
            'instruments': [],
            'locations': [],
            'location_types': [],
            'objects': [],
            'action_types': []
        }

        for share in shares:
            obj = shared_object_preprocessor(share.object_id, share.policy, refs, markdown_images, sharing_user_id=share.user_id)
            result_lists['objects'].append(obj)

        while len(refs) > 0:
            type, id = refs.pop()
            if type == 'instruments' and flask.current_app.config['DISABLE_INSTRUMENTS']:
                continue
            if type in ref_ids and id not in ref_ids[type]:
                processed = preprocessors[type](id, component, refs, markdown_images)
                if processed is not None:
                    result_lists[type].append(processed)
                ref_ids[type].append(id)

        return {
            'header': _get_header(component),
            'markdown_images': markdown_images,
            'actions': result_lists['actions'],
            'users': result_lists['users'],
            'instruments': result_lists['instruments'],
            'locations': result_lists['locations'],
            'location_types': result_lists['location_types'],
            'objects': result_lists['objects'],
            'action_types': result_lists['action_types']
        }


class Users(Resource):
    @http_token_auth.login_required
    def get(self) -> ResponseData:
        last_sync = _get_last_sync(flask.request.args)
        component = flask.g.component

        result_users = []
        federation_candidates = []

        all_users = get_users()
        user_by_id: dict[int, User] = {
            user.id: user
            for user in all_users
        }

        shared_user_aliases = get_user_aliases_for_component(component.id, modified_since=last_sync)

        for candidate in all_users:
            if candidate.type == UserType.PERSON:
                federation_candidates.append({
                    'user_id': candidate.id,
                    'email_hashes': get_user_email_hashes(candidate)
                })

        for alias in shared_user_aliases:
            user = user_by_id.get(alias.user_id)
            if user is None:
                continue
            if user.component_id is not None:
                continue
            if user.is_hidden or not user.is_active:
                continue
            result_users.append(
                {
                    'user_id': alias.user_id,
                    'component_uuid': flask.current_app.config['FEDERATION_UUID'],
                    'name': alias.name,
                    'email': alias.email,
                    'orcid': alias.orcid,
                    'affiliation': alias.affiliation,
                    'role': alias.role,
                    'extra_fields': alias.extra_fields
                }
            )

        return {
            'header': _get_header(component),
            'users': result_users,
            'federation_candidates': federation_candidates
        }


class File(Resource):
    @http_token_auth.login_required
    def get(self, object_id: int, file_id: int) -> ResponseData:
        component = flask.g.component

        try:
            share = get_share(object_id, component.id)
            if 'access' not in share.policy or not share.policy['access'].get('files'):
                return {
                    "message": f"files linked to object {object_id} are not shared"
                }, 403
        except errors.ObjectDoesNotExistError:
            return {
                "message": f"object {object_id} does not exist"
            }, 404
        except errors.ShareDoesNotExistError:
            return {
                "message": f"object {object_id} is not shared"
            }, 403

        try:
            file = get_file(file_id, object_id)
        except errors.FileDoesNotExistError:
            return {
                "message": f"file {file_id} of object {object_id} does not exist"
            }, 404

        if file.storage == 'database':
            try:
                with file.open() as f:
                    response = flask.make_response(f.read())
                    response.headers.set('Content-Disposition', 'attachment', filename=file.original_file_name)
                    return response
            except Exception:
                return {
                    "message": f"file {file_id} of object {object_id} could not be read"
                }, 404
        elif file.storage == 'url':
            file_json = {
                'header': _get_header(component),
                'object_id': file.object_id,
                'file_id': file.id,
                'storage': file.storage,
                'url': file.url
            }
            return file_json, 200

        return {
            "message": f"file {file_id} of object {object_id} is not shareable"
        }, 403


class Components(Resource):
    @http_token_auth.login_required
    def get(self) -> ResponseData:
        return {
            'header': _get_header(flask.g.component),
            'discoverable': flask.current_app.config['ENABLE_FEDERATION_DISCOVERABILITY'],
            'components': [
                {
                    'name': component.name,
                    'uuid': component.uuid,
                    'source_uuid': flask.current_app.config['FEDERATION_UUID'],
                    'address': component.address,
                    'distance': 1,
                    'discoverable': component.discoverable
                }
                for component in get_components()
            ] + [
                {
                    'name': component_info.name,
                    'uuid': component_info.uuid,
                    'source_uuid': component_info.source_uuid,
                    'address': component_info.address,
                    'distance': component_info.distance + 1,
                    'discoverable': component_info.discoverable
                }
                for component_info in get_component_infos()
            ]
        }


class FederatedLoginMetadata(Resource):
    @http_token_auth.login_required
    def get(self) -> ResponseData:
        if flask.current_app.config['ENABLE_FEDERATED_LOGIN']:
            return {
                'idp': get_idp_metadata().decode('utf-8'),
                'sp': get_sp_metadata(flask.g.component).decode('utf-8'),
                'enabled': check_component_locally_ready(flask.g.component)
            }
        else:
            return {}



# coding: utf-8
"""

"""

import flask
import flask_login
from flask_babel import lazy_gettext
from flask_wtf import FlaskForm
from wtforms import SelectField, IntegerField, FloatField, TextAreaField, HiddenField, FileField, StringField, BooleanField, RadioField
from wtforms.validators import InputRequired, ValidationError, Optional

from ...logic import errors
from ...models import Permissions, User
from ...logic.publications import simplify_doi
from ...logic.errors import InvalidDOIError
from ...logic.utils import get_translated_text
from ...logic.groups import get_groups
from ...logic.users import get_users
from ...logic.projects import get_projects
from ..validators import ObjectIdValidator
from ...logic.utils import parse_url


class ObjectForm(FlaskForm):
    pass


class ObjectVersionRestoreForm(FlaskForm):
    pass


class CommentForm(FlaskForm):
    content = TextAreaField(validators=[InputRequired()])


class FileForm(FlaskForm):
    file_source = HiddenField(validators=[InputRequired()])
    file_names = HiddenField()
    local_files = FileField()

    def validate_file_source(form, field: StringField) -> None:
        if field.data not in ['local']:
            raise ValidationError('Invalid file source')


def _validate_url(url: str) -> None:
    try:
        parse_url(url)
    except errors.InvalidURLError:
        raise ValidationError(0)
    except errors.URLTooLongError:
        raise ValidationError(1)
    except errors.InvalidIPAddressError:
        raise ValidationError(2)
    except errors.InvalidPortNumberError:
        raise ValidationError(3)


class ExternalLinkForm(FlaskForm):
    url = StringField()

    def validate_url(form, field: StringField) -> None:
        _validate_url(field.data)


class FileInformationForm(FlaskForm):
    title = StringField()
    url = StringField()
    description = TextAreaField()

    def validate_url(form, field: StringField) -> None:
        if field.data is not None:
            _validate_url(field.data)


class FileHidingForm(FlaskForm):
    reason = TextAreaField()


class ObjectLocationAssignmentForm(FlaskForm):
    location = SelectField(validators=[InputRequired()])
    responsible_user = SelectField(validators=[InputRequired()])
    description = StringField()


class ObjectPublicationForm(FlaskForm):
    doi = StringField()
    title = StringField()
    object_name = StringField()

    def validate_doi(form, field: StringField) -> None:
        try:
            field.data = simplify_doi(field.data)
        except InvalidDOIError:
            raise ValidationError('Please enter a valid DOI')


class CopyPermissionsForm(FlaskForm):
    object_id = SelectField(validators=[ObjectIdValidator(Permissions.GRANT), InputRequired()], validate_choice=False)


class ObjectNewShareAccessForm(FlaskForm):
    component_id = IntegerField(validators=[InputRequired()])

    data = BooleanField()
    action = BooleanField()
    users = BooleanField()
    files = BooleanField()
    comments = BooleanField()
    object_location_assignments = BooleanField()


class ObjectEditShareAccessForm(FlaskForm):
    component_id = IntegerField(validators=[InputRequired()])

    data = BooleanField()
    action = BooleanField()
    users = BooleanField()
    files = BooleanField()
    comments = BooleanField()
    object_location_assignments = BooleanField()


class UseInActionForm(FlaskForm):
    action_type_id = HiddenField()
    action_id = HiddenField()
    objects = HiddenField(validators=[InputRequired()])


class GenerateLabelsForm(FlaskForm):
    form_variant = SelectField(validators=[InputRequired()], choices=[('mixed-formats', lazy_gettext('Mixed label formats')), ('fixed-width', lazy_gettext('Fixed-width labels')), ('minimal-height', lazy_gettext('Minimal-height labels'))])
    objects = HiddenField(validators=[InputRequired()])
    paper_size = SelectField(validators=[InputRequired()], choices=[("DIN A4 (Portrait)", lazy_gettext('DIN A4 (Portrait)')), ("DIN A4 (Landscape)", lazy_gettext('DIN A4 (Landscape)')), ("Letter (Portrait)", lazy_gettext('Letter / ANSI A (Portrait)')), ("Letter (Landscape)", lazy_gettext('Letter / ANSI A (Landscape)'))])
    label_width = FloatField()
    min_label_width = FloatField()
    min_label_height = FloatField()
    labels_per_object = IntegerField()
    center_qr_ghs = BooleanField()
    qr_ghs_side_by_side = BooleanField()
    include_qr = BooleanField()


class EditPermissionsForm(FlaskForm):
    objects = HiddenField(validators=[InputRequired()])
    target_type = SelectField(validators=[InputRequired()])
    update_mode = RadioField(validators=[InputRequired()], choices=[('set-min', lazy_gettext('Set minimum Permissions')), ('set-max', lazy_gettext('Set maximum Permissions'))])
    groups = SelectField(validators=[Optional()])
    project_groups = SelectField(validators=[Optional()])
    users = SelectField(validators=[Optional()])
    permission = RadioField(validators=[InputRequired()], choices=[
                            (Permissions.NONE.name.lower(), lazy_gettext('None')),
                            (Permissions.READ.name.lower(), lazy_gettext('Read')),
                            (Permissions.WRITE.name.lower(), lazy_gettext('Write')),
                            (Permissions.GRANT.name.lower(), lazy_gettext('Grant'))
                            ])

    def __init__(self) -> None:
        super().__init__()
        target_type_choices = [
            ('signed-in-users', lazy_gettext('All Signed-In Users')),
            ('group', lazy_gettext('Basic Group')),
            ('project-group', lazy_gettext('Project Group')),
            ('user', lazy_gettext('User'))
        ]
        if flask.current_app.config["ENABLE_ANONYMOUS_USERS"]:
            target_type_choices.insert(1, ('anonymous', lazy_gettext('Anonymous Users')))

        self.target_type.choices = target_type_choices
        self.groups.choices = [(group.id, get_translated_text(group.name)) for group in get_groups()]
        self.project_groups.choices = [(project.id, get_translated_text(project.name)) for project in get_projects()]
        self.users.choices = [(user.id, f"{user.name} (#{user.id})") for user in get_users(order_by=User.id, exclude_fed=True, exclude_eln_import=True, exclude_hidden=not flask_login.current_user.is_admin or not flask_login.current_user.settings['SHOW_HIDDEN_USERS_AS_ADMIN'])]



import typing

import flask

from ... import db, logic
from . import core
from ...models import BackgroundTask, BackgroundTaskStatus, UpdatableObjectsCheckStatus, AutomaticSchemaUpdateStatus, Objects
from ..automatic_schema_updates import get_updatable_objects_check, set_updatable_objects_check_status, get_automatic_schema_update, set_automatic_schema_update_status


def post_check_for_automatic_schema_updates_task(
        *,
        updatable_objects_check_id: int
) -> typing.Tuple[BackgroundTaskStatus, typing.Optional[BackgroundTask]]:
    return core.post_background_task(
        type='check_for_automatic_schema_updates',
        data={
            'updatable_objects_check_id': updatable_objects_check_id
        },
        auto_delete=False
    )


def post_perform_automatic_schema_updates_task(
        automatic_schema_update_id: int,
        auto_delete: bool = True
) -> typing.Tuple[BackgroundTaskStatus, typing.Optional[BackgroundTask]]:
    return core.post_background_task(
        type='perform_automatic_schema_updates',
        data={
            'automatic_schema_update_id': automatic_schema_update_id,
        },
        auto_delete=auto_delete
    )


def handle_check_for_automatic_schema_updates_task(
        data: typing.Dict[str, typing.Any],
        task_id: typing.Optional[int]
) -> typing.Tuple[bool, typing.Optional[dict[str, typing.Any]]]:
    updatable_objects_check_id = typing.cast(int, data['updatable_objects_check_id'])
    updatable_objects_check = get_updatable_objects_check(updatable_objects_check_id)
    previous_object_id = -1
    automatically_updatable_objects = []
    manually_updatable_objects = []
    if updatable_objects_check.status == UpdatableObjectsCheckStatus.POSTED:
        set_updatable_objects_check_status(
            updatable_objects_check_id=updatable_objects_check_id,
            status=UpdatableObjectsCheckStatus.IN_PROGRESS,
            result={
                'previous_object_id': previous_object_id,
                'automatically_updatable_objects': [],
                'manually_updatable_objects': [],
            }
        )
    elif updatable_objects_check.status == UpdatableObjectsCheckStatus.IN_PROGRESS:
        if updatable_objects_check.result:
            previous_object_id = updatable_objects_check.result.get('previous_object_id', -1)
            automatically_updatable_objects = updatable_objects_check.result.get('automatically_updatable_objects', [])
            manually_updatable_objects = updatable_objects_check.result.get('manually_updatable_objects', [])
    else:
        return True, None
    num_objects_to_handle = 100 if flask.current_app.config['ENABLE_BACKGROUND_TASKS'] else None
    num_objects_found_list = [0]

    objects = logic.objects.get_objects(
        filter_func=lambda data: db.and_(
            db.and_(
                Objects.current_table.c.action_id.in_(updatable_objects_check.action_ids) if updatable_objects_check.action_ids else db.true(),
                Objects.current_table.c.object_id > previous_object_id,
            ),
            db.and_(
                Objects.current_table.c.fed_object_id == db.null(),
                Objects.current_table.c.eln_object_id == db.null()
            )
        ),
        sorting_func=lambda current_columns, original_columns: db.sql.asc(current_columns.object_id),
        limit=num_objects_to_handle,
        num_objects_found=num_objects_found_list,
    )
    cached_actions = {}
    for object in objects:
        if object.data is None:
            continue
        if object.schema is None:
            continue
        if object.action_id is None:
            continue
        if object.action_id not in cached_actions:
            cached_actions[object.action_id] = logic.actions.get_action(action_id=object.action_id)
        action = cached_actions[object.action_id]
        if action.schema is None:
            continue
        if action.schema != object.schema:
            converting_necessary, messages = logic.schemas.is_converting_to_schema_necessary(object.data, object.schema, action.schema)
            if converting_necessary:
                manually_updatable_objects.append((object.object_id, object.version_id, messages))
            else:
                automatically_updatable_objects.append((object.object_id, object.version_id, messages))
    num_objects_found = num_objects_found_list[0]
    set_updatable_objects_check_status(
        updatable_objects_check_id=updatable_objects_check_id,
        status=UpdatableObjectsCheckStatus.IN_PROGRESS if num_objects_to_handle and num_objects_found > num_objects_to_handle else UpdatableObjectsCheckStatus.DONE,
        result={
            'previous_object_id': max(
                object.object_id
                for object in objects
            ),
            'automatically_updatable_objects': automatically_updatable_objects,
            'manually_updatable_objects': manually_updatable_objects
        }
    )
    if num_objects_to_handle and num_objects_found > num_objects_to_handle:
        post_check_for_automatic_schema_updates_task(
            updatable_objects_check_id=updatable_objects_check_id
        )
    return True, None


def handle_perform_automatic_schema_updates_task(
        data: typing.Dict[str, typing.Any],
        task_id: typing.Optional[int]
) -> typing.Tuple[bool, typing.Optional[dict[str, typing.Any]]]:
    automatic_schema_update_id = typing.cast(int, data['automatic_schema_update_id'])
    automatic_schema_update = get_automatic_schema_update(automatic_schema_update_id=automatic_schema_update_id)
    object_ids = automatic_schema_update.object_ids
    object_ids_success: typing.List[int] = []
    object_ids_failure: typing.List[int] = []
    if automatic_schema_update.status == AutomaticSchemaUpdateStatus.POSTED:
        set_automatic_schema_update_status(
            automatic_schema_update_id=automatic_schema_update_id,
            status=AutomaticSchemaUpdateStatus.IN_PROGRESS,
            result={
                'object_ids_success': object_ids_success,
                'object_ids_failure': object_ids_failure,
            }
        )
    elif automatic_schema_update.status == AutomaticSchemaUpdateStatus.IN_PROGRESS:
        if automatic_schema_update.result:
            object_ids_success = automatic_schema_update.result.get('object_ids_success', object_ids_success)
            object_ids_failure = automatic_schema_update.result.get('object_ids_failure', object_ids_failure)
    else:
        return True, None

    if object_ids_success or object_ids_failure:
        object_ids = [
            object_id
            for object_id in object_ids
            if object_id not in object_ids_success and object_id not in object_ids_failure
        ]
    if not object_ids:
        set_automatic_schema_update_status(
            automatic_schema_update_id=automatic_schema_update_id,
            status=AutomaticSchemaUpdateStatus.DONE,
            result=automatic_schema_update.result,
        )
        return True, None
    num_objects_to_handle = 100 if flask.current_app.config['ENABLE_BACKGROUND_TASKS'] else None

    if num_objects_to_handle is None:
        unhandled_object_ids = []
    else:
        unhandled_object_ids = object_ids[num_objects_to_handle:]
        object_ids = object_ids[:num_objects_to_handle]
    objects = logic.objects.get_objects(
        filter_func=lambda data: db.and_(
            Objects.current_table.c.object_id.in_(object_ids),
            db.and_(
                Objects.current_table.c.fed_object_id == db.null(),
                Objects.current_table.c.eln_object_id == db.null()
            )
        ),
        sorting_func=lambda current_columns, original_columns: db.sql.asc(current_columns.object_id),
    )
    cached_actions = {}
    for object in objects:
        if object.data is None:
            object_ids_failure.append(object.object_id)
            continue
        if object.schema is None:
            object_ids_failure.append(object.object_id)
            continue
        if object.action_id is None:
            object_ids_failure.append(object.object_id)
            continue
        if object.action_id not in cached_actions:
            cached_actions[object.action_id] = logic.actions.get_action(action_id=object.action_id)
        action = cached_actions[object.action_id]
        if action.schema is None:
            object_ids_failure.append(object.object_id)
            continue
        if action.schema != object.schema:
            converting_to_schema_necessary, conversion_messages = logic.schemas.is_converting_to_schema_necessary(object.data, object.schema, action.schema)
            if converting_to_schema_necessary or conversion_messages:
                object_ids_failure.append(object.object_id)
                continue
            try:
                logic.objects.update_object(
                    object_id=object.object_id,
                    data=object.data,
                    schema=action.schema,
                    user_id=automatic_schema_update.user_id,
                )
                object_ids_success.append(object.object_id)
            except Exception:
                object_ids_failure.append(object.object_id)
    set_automatic_schema_update_status(
        automatic_schema_update_id=automatic_schema_update_id,
        status=AutomaticSchemaUpdateStatus.IN_PROGRESS if unhandled_object_ids else AutomaticSchemaUpdateStatus.DONE,
        result={
            'object_ids_success': object_ids_success,
            'object_ids_failure': object_ids_failure
        }
    )
    if unhandled_object_ids:
        post_perform_automatic_schema_updates_task(
            automatic_schema_update_id=automatic_schema_update_id
        )
    return True, None



# coding: utf-8
"""
Implementation of OpenID Connect.
"""

import hashlib
import secrets
import time
import typing

import flask
import pydantic
from furl import furl
from simple_openid_connect.pkce import generate_pkce_pair
from simple_openid_connect.client import OpenidClient
from simple_openid_connect.data import (
    IdToken, JwtAccessToken, RpInitiatedLogoutRequest, TokenErrorResponse,
    TokenIntrospectionSuccessResponse, TokenSuccessResponse,
    UserinfoErrorResponse
)
from simple_openid_connect.exceptions import UnsupportedByProviderError

from ..config import OIDC_REQUIRED_CONFIG_KEYS
from ..models import AuthenticationType, UserType
from . import authentication, users


def is_oidc_configured() -> bool:
    """
    Check whether all required OIDC configuration values are set.

    :return: whether all values are set, False otherwise
    """
    return all(
        flask.current_app.config[key]
        for key in OIDC_REQUIRED_CONFIG_KEYS
    )


def is_oidc_only_auth_method() -> bool:
    """
    Check whether OIDC is configured and OIDC and api keys are the only
    permitted authentication method.

    :return: whether oidc is the only available auth method, False otherwise
    """
    return is_oidc_configured() and flask.current_app.config['OIDC_ONLY']


def init_app(app: flask.Flask) -> None:
    """
    Initialize OIDC client. Requires valid OIDC config and app or request
    context.
    """
    client = OpenidClient.from_issuer_url(
        app.config['OIDC_ISSUER'],
        app.url_for('frontend.oidc_callback', _external=True),
        app.config['OIDC_CLIENT_ID'],
        app.config['OIDC_CLIENT_SECRET'],
        app.config['OIDC_SCOPES'],
    )
    app.extensions['sampledb.logic.oidc.client'] = client
    app.extensions['sampledb.logic.oidc.data'] = {}


def _get_client() -> OpenidClient:
    return typing.cast(
        OpenidClient,
        flask.current_app.extensions['sampledb.logic.oidc.client']
    )


class _Data(pydantic.BaseModel):
    redirect_uri: str
    state: str
    nonce_value: str | None
    code_challenge_method: typing.Literal['S256'] | None
    code_verifier: str | None
    started: float = pydantic.Field(default_factory=time.time)


def start_authentication(redirect_uri: str) -> typing.Tuple[str, str]:
    """
    Start an OIDC authentication attempt.

    The caller must redirect to the returned url and store the returned opaque
    string token in the user session by setting it via a HttpOnly cookie. The
    token must be given as an argument to `handle_authentication` and removed
    from the user session.

    :param redirect_uri: Where the user should be redirected to after the
        authentication has been finished.
    :return: A tuple of the url and the token.
    """

    client = _get_client()

    state = secrets.token_urlsafe(32)

    if flask.current_app.config['OIDC_DISABLE_NONCE']:
        nonce_value = None
        nonce_hash = None
    else:
        nonce_value = secrets.token_urlsafe(32)
        nonce_hash = hashlib.sha256(nonce_value.encode()).hexdigest()

    code_challenge_method: typing.Optional[typing.Literal['S256']]
    if 'S256' in getattr(client.provider_config, 'code_challenge_methods_supported', []):
        code_challenge_method = 'S256'
        code_verifier, code_challenge = generate_pkce_pair()
    else:
        code_challenge_method = None
        code_verifier = None
        code_challenge = None

    url = client.authorization_code_flow.start_authentication(
        state=state,
        nonce=nonce_hash,
        code_challenge_method=code_challenge_method,
        code_challenge=code_challenge,
    )
    data = _Data(
        redirect_uri=redirect_uri,
        state=state,
        nonce_value=nonce_value,
        code_challenge_method=code_challenge_method,
        code_verifier=code_verifier,
    ).model_dump_json()
    return url, data


def handle_authentication(url: str, token: str) -> tuple[users.User, str, str]:
    """
    Handle an OIDC authentication callback.

    The caller must provide the entire callback URL and the token returned by
    `start_authentication`. The token must then be removed from the user
    session.

    The function will either return the user, the ID token, and the next URL,
    or raise an error. The ID token should be saved and passed to the logout
    function. The next URL is the address the user tried to visit before being
    asked to authenticate.

    :param url: The request url.
    :param token: The token returned by `start_authentication`.
    :return: A tuple of the user, the ID token, and the next URL.
    """

    current_url = furl(url)
    data = _Data.model_validate_json(token)

    if current_url.args['state'] != data.state:
        raise RuntimeError('Incorrect state value')

    # Make the timeout configurable if necessary.
    if data.started < time.time() - 10 * 60:
        raise RuntimeError('Request timed out')

    client = _get_client()

    token_response = client.authorization_code_flow.handle_authentication_result(
        url,
        state=data.state,
        code_verifier=data.code_verifier,
    )

    if isinstance(token_response, TokenErrorResponse):
        raise RuntimeError('Token Error')

    if data.nonce_value is not None:
        nonce = hashlib.sha256(data.nonce_value.encode()).hexdigest()
    else:
        nonce = None

    id_token = client.decode_id_token(
        token_response.id_token,
        nonce=nonce,
    )

    authentication_methods = authentication.get_oidc_authentications_by_sub(
        id_token.iss,
        id_token.sub,
    )

    if len(authentication_methods) > 1:
        raise RuntimeError(
            f'More than one user with iss = "{id_token.iss}" '
            f'and sub = "{id_token.sub}"'
        )

    if authentication_methods:
        user = users.get_user(authentication_methods[0].user_id)
    else:
        user = _handle_new_user(client, token_response, id_token)

    user = _update_user(
        user=user,
        id_token=id_token,
    )

    # Do not handle errors, see note in function.
    user = _handle_roles(user, id_token)

    return user, token_response.id_token, data.redirect_uri


def logout(id_token_hint: str, post_logout_redirect_uri: str) -> str:
    """
    Return the URL to initiate a logout request with the OIDC Provider. If the
    OP does not support RP-Initiated Logout, the returned URL will default to
    the `post_logout_redirect_uri`.

    In either case, the user will ultimately be redirected to the
    `post_logout_redirect_uri`, whether by the OP or by the caller of this
    function.

    :param id_token_hint: The ID token returned in the authentication.
    :param post_logout_redirect_uri: The URL to which the OP should redirect
        the user.
    :return: The next URL.
    """
    client = _get_client()
    try:
        return client.initiate_logout(RpInitiatedLogoutRequest(
            id_token_hint=id_token_hint,
            client_id=client.client_auth.client_id,
            post_logout_redirect_uri=post_logout_redirect_uri,
        ))
    except UnsupportedByProviderError:
        return post_logout_redirect_uri


def validate_access_token(access_token: str) -> typing.Tuple[authentication.AuthenticationMethod, users.User]:
    """
    Validate the access token and return the corresponding authentication
    method and user, or raise an error if it's not a valid token.

    This does not create or update user accounts, aside from the OIDC roles.
    Those will be updated and their omission will fail the validation.

    Depending on the access token, a network request may be necessary to
    validate the token. OIDC_ACCESS_TOKEN_ALLOW_INTROSPECTION can be used to
    permit or deny making that request.

    :param access_token: The Access Token.
    :return: A tuple of the authentication method and the user.
    """

    response = _validate_access_token(access_token)

    if not response.iss or not response.sub:
        raise RuntimeError('Missing iss or sub.')

    [authentication_method] = authentication.get_oidc_authentications_by_sub(
        response.iss,
        response.sub,
    )

    user = users.get_user(authentication_method.user_id)
    # Do not handle errors, see note in function.
    user = _handle_roles(user, response)

    return authentication_method, user


def _validate_access_token(access_token: str) -> typing.Union[JwtAccessToken, TokenIntrospectionSuccessResponse]:
    client = _get_client()

    # Try to parse the token as a JwtAccessToken. If it's a JWT, validate it
    # and return the result or bail out if the validation fails.
    # If it's not a JWT and the config permits token introspection, use the
    # Introspection Endpoint and return a valid response or bail out.
    try:
        token = JwtAccessToken.parse_jwt(access_token, client.provider_keys)
    except Exception:
        pass
    else:
        token.validate_extern(
            client.provider_config.issuer,
            client.client_auth.client_id,
        )
        return token

    if flask.current_app.config['OIDC_ACCESS_TOKEN_ALLOW_INTROSPECTION']:
        introspection_response = client.introspect_token(
            access_token,
            token_type_hint='access_token',
        )
        if isinstance(introspection_response, TokenIntrospectionSuccessResponse):
            return introspection_response
        raise RuntimeError('Introspection Endpoint returned an error')

    raise RuntimeError('Not a JWT and not allowed to use introspection')


def _update_user(user: users.User, id_token: IdToken) -> users.User:
    name = typing.cast(
        typing.Union[str, None],
        getattr(id_token, 'name', None),
    )
    email = typing.cast(
        typing.Union[str, None],
        getattr(id_token, 'email', None),
    )
    email_verified = typing.cast(
        typing.Union[bool, None],
        getattr(id_token, 'email_verified', None),
    )

    if (
        name is not None and
        flask.current_app.config['ENFORCE_SPLIT_NAMES'] and
        ', ' not in name[1:-1]
    ):
        name = user.name

    if (
            name is not None and
            email is not None and
            email_verified and
            (name != user.name or email != user.email)
    ):
        users.update_user(
            user.id,
            updating_user_id=None,
            name=name,
            email=email,
        )
        return users.get_user(user.id)

    return user


def _handle_new_user(client: OpenidClient, token_response: TokenSuccessResponse, id_token: IdToken) -> users.User:
    if flask.current_app.config['OIDC_CREATE_ACCOUNT'] == 'no':
        raise RuntimeError('Automatic account creation disabled.')

    userinfo = client.fetch_userinfo(token_response.access_token)

    if isinstance(userinfo, UserinfoErrorResponse):
        raise RuntimeError('Userinfo error')

    try:
        name = typing.cast(str, userinfo.name)  # type: ignore
        email = typing.cast(str, userinfo.email)  # type: ignore
        email_verified = typing.cast(
            str, userinfo.email_verified)  # type: ignore
        if not name or '@' not in email:
            raise RuntimeError()
    except Exception as e:
        raise RuntimeError('Missing required attributes in userinfo') from e

    if not email_verified:
        raise RuntimeError('Email not verified')

    matching_users = users.get_users_by_email(email)

    if len(matching_users) > 1:
        raise RuntimeError('Multiple matching accounts.')
    if len(matching_users) == 1:
        if flask.current_app.config['OIDC_CREATE_ACCOUNT'] == 'deny_existing':
            raise RuntimeError('Automatic linking is disabled.')

        authentication_methods = authentication.get_authentication_methods(
            matching_users[0].id,
            {AuthenticationType.OIDC}
        )
        if authentication_methods:
            if any(method.login['iss'] == id_token.iss for method in authentication_methods):
                raise RuntimeError('Unstable subject identifier from issuer')
            raise RuntimeError('OIDC authentication method from other issuer')

        user = matching_users[0]
    else:
        user = users.create_user(
            name=name,
            email=email,
            type=UserType.PERSON,
        )

    authentication.add_oidc_authentication(
        user.id,
        id_token.iss,
        id_token.sub,
    )

    return user


def _handle_roles(
        user: users.User,
        obj: typing.Union[IdToken, JwtAccessToken, TokenIntrospectionSuccessResponse],
) -> users.User:
    if not flask.current_app.config['OIDC_ROLES']:
        return user

    # Note: Intentionally fail on any error. It's probably a config error and
    # we do not want anyone to login with incorrect permissions.

    i = obj.model_dump()
    for path in flask.current_app.config['OIDC_ROLES'].split('.'):
        if isinstance(i, list):
            path = int(path)
        i = i[path]  # pyright: ignore
    roles = typing.cast(list[str], i)

    def f(prop: str, func: typing.Callable[[int, bool], None]) -> bool:
        v = None
        if f'is_{prop}' in roles:
            v = True
        if f'is_not_{prop}' in roles:
            v = False
        if v is not None and v != getattr(user, f'is_{prop}'):
            func(user.id, v)
            return True
        return False

    refresh = False
    refresh |= f('active', users.set_user_active)
    refresh |= f('hidden', users.set_user_hidden)
    refresh |= f('readonly', users.set_user_readonly)
    refresh |= f('admin', users.set_user_administrator)
    if refresh:
        user = users.get_user(user.id)
    return user




from . import core
from . import send_mail
from . import background_dataverse_export
from .automatic_schema_updates import post_check_for_automatic_schema_updates_task, post_perform_automatic_schema_updates_task
from .core import start_handler_threads, stop_handler_threads, post_background_task, get_background_tasks, get_background_task_result, reset_claimed_background_tasks
from .send_mail import post_send_mail_task
from .background_dataverse_export import post_dataverse_export_task
from .poke_components import post_poke_components_task
from .trigger_webhooks import post_trigger_object_log_webhooks

__all__ = [
    'core',
    'send_mail',
    'background_dataverse_export',
    'post_send_mail_task',
    'start_handler_threads',
    'stop_handler_threads',
    'post_background_task',
    'get_background_tasks',
    'get_background_task_result',
    'post_check_for_automatic_schema_updates_task',
    'post_dataverse_export_task',
    'post_perform_automatic_schema_updates_task',
    'post_poke_components_task',
    'post_trigger_object_log_webhooks',
    'reset_claimed_background_tasks',
]



# coding: utf-8
"""
Logic module handling communication with other components in a SampleDB federation
"""
import typing
import datetime
import io
import flask
import requests

from .utils import _get_dict, _get_list, _get_bool
from .users import import_user, parse_user
from .components import import_component_info, parse_component_info
from .location_types import import_location_type, parse_location_type
from .instruments import import_instrument, parse_instrument
from .action_types import import_action_type, parse_action_type
from .locations import import_location, parse_location, locations_check_for_cyclic_dependencies
from .markdown_images import parse_markdown_image, import_markdown_image
from .actions import import_action, parse_action
from .objects import import_object, parse_object
from ..components import Component, set_component_discoverable
from ..component_authentication import get_own_authentication
from ..users import link_users_by_email_hashes
from ..federation.login import update_metadata
from .. import errors
from ...models import ComponentAuthenticationType

PROTOCOL_VERSION_MAJOR = 0
PROTOCOL_VERSION_MINOR = 1

FEDERATION_TIMEOUT = 60


def _send_request(
        method: typing.Union[typing.Literal['get'], typing.Literal['post'], typing.Literal['put']],
        endpoint: str,
        component: Component,
        headers: typing.Optional[typing.Dict[str, str]] = None,
        **kwargs: typing.Any
) -> requests.Response:
    if component.address is None:
        raise errors.MissingComponentAddressError()
    if headers is None:
        headers = {}
    auth = get_own_authentication(component.id, ComponentAuthenticationType.TOKEN)

    if auth:
        headers['Authorization'] = 'Bearer ' + auth.login['token']

    url = component.address.rstrip('/') + endpoint

    method_callable = {
        'get': requests.get,
        'post': requests.post,
        'put': requests.put,
    }[method]
    return method_callable(  # type: ignore
        url,
        headers=headers,
        timeout=FEDERATION_TIMEOUT,
        **kwargs
    )


def post(
        endpoint: str,
        component: Component,
        payload: typing.Optional[typing.Dict[str, typing.Any]] = None,
        headers: typing.Optional[typing.Dict[str, str]] = None
) -> None:
    _send_request('post', endpoint, component, headers, data=payload)


def put(
        endpoint: str,
        component: Component,
        headers: typing.Optional[typing.Dict[str, str]] = None,
        **kwargs: typing.Any
) -> None:
    _send_request('put', endpoint, component, headers, **kwargs)


def get_binary(
    endpoint: str,
    component: Component,
    headers: typing.Optional[typing.Dict[str, str]] = None
) -> typing.BinaryIO:
    response = _send_request('get', endpoint, component, headers)
    if response.status_code == 401:
        # 401 Unauthorized
        raise errors.UnauthorizedRequestError()
    if response.status_code in [500, 501, 502, 503, 504]:
        raise errors.RequestServerError()
    if response.status_code != 200:
        raise errors.RequestError()
    return io.BytesIO(response.content)


def get(
        endpoint: str,
        component: Component,
        headers: typing.Optional[typing.Dict[str, str]] = None,
        *,
        ignore_last_sync_time: bool = False
) -> typing.Dict[str, typing.Any]:
    if component.address is None:
        raise errors.MissingComponentAddressError()
    if headers is None:
        headers = {}
    auth = get_own_authentication(component.id, ComponentAuthenticationType.TOKEN)

    if auth:
        headers['Authorization'] = 'Bearer ' + auth.login['token']

    parameters = {}
    if component.last_sync_timestamp is not None and not ignore_last_sync_time:
        parameters['last_sync_timestamp'] = component.last_sync_timestamp.strftime('%Y-%m-%d %H:%M:%S.%f')
    req = requests.get(
        component.address.rstrip('/') + endpoint,
        headers=headers,
        params=parameters,
        timeout=FEDERATION_TIMEOUT
    )
    if req.status_code == 401:
        # 401 Unauthorized
        raise errors.UnauthorizedRequestError()
    if req.status_code in [500, 501, 502, 503, 504]:
        raise errors.RequestServerError()
    if req.status_code != 200:
        raise errors.RequestError()
    try:
        return req.json()  # type: ignore
    except ValueError:
        raise errors.InvalidJSONError()


def update_poke_component(
        component: Component
) -> None:
    post('/federation/v1/hooks/update/', component)


def _validate_header(
        header: typing.Optional[typing.Dict[str, typing.Any]],
        component: Component
) -> None:
    if header is not None:
        if header.get('db_uuid') != component.uuid:
            raise errors.InvalidDataExportError(f'UUID of exporting database ({header.get("db_uuid")}) does not match expected UUID ({component.uuid}).')
        protocol_version = header.get('protocol_version')
        if protocol_version is not None:
            if 'major' not in protocol_version or 'minor' not in protocol_version:
                raise errors.InvalidDataExportError(f'Invalid protocol version \'{header.get("protocol_version")}\'')
            try:
                major, minor = int(protocol_version['major']), int(protocol_version['minor'])
                if major > PROTOCOL_VERSION_MAJOR or (major <= PROTOCOL_VERSION_MAJOR and minor > PROTOCOL_VERSION_MINOR):
                    raise errors.InvalidDataExportError(f'Unsupported protocol version {header.get("protocol_version")}')
            except ValueError:
                raise errors.InvalidDataExportError(f'Invalid protocol version {header.get("protocol_version")}')
        else:
            raise errors.InvalidDataExportError('Missing protocol_version.')


def import_updates(
        component: Component,
        *,
        ignore_last_sync_time: bool = False
) -> None:
    if flask.current_app.config['FEDERATION_UUID'] is None:
        raise errors.ComponentNotConfiguredForFederationError()
    timestamp = datetime.datetime.now(datetime.timezone.utc)
    components = None
    try:
        components = get('/federation/v1/shares/components/', component, ignore_last_sync_time=ignore_last_sync_time)
    except errors.InvalidJSONError:
        raise errors.InvalidDataExportError('Received an invalid JSON string.')
    except errors.RequestServerError:
        pass
    except errors.UnauthorizedRequestError:
        pass
    if components:
        update_components(component, components)
    users = None
    try:
        users = get('/federation/v1/shares/users/', component, ignore_last_sync_time=ignore_last_sync_time)
    except errors.InvalidJSONError:
        raise errors.InvalidDataExportError('Received an invalid JSON string.')
    except errors.RequestServerError:
        pass
    except errors.UnauthorizedRequestError:
        pass
    if users:
        update_users(component, users)
    try:
        updates = get('/federation/v1/shares/objects/', component, ignore_last_sync_time=ignore_last_sync_time)
    except errors.InvalidJSONError:
        raise errors.InvalidDataExportError('Received an invalid JSON string.')
    if updates:
        update_shares(component, updates)

    metadata = get('/federation/v1/shares/metadata/', component)
    if metadata is not None:
        update_metadata(component, metadata)
    component.update_last_sync_timestamp(timestamp)


def update_components(
        component: Component,
        updates: typing.Dict[str, typing.Any]
) -> None:
    _get_dict(updates, mandatory=True)
    _validate_header(updates.get('header'), component)
    discoverable = _get_bool(updates.get('discoverable'), mandatory=True)
    set_component_discoverable(component.id, discoverable=discoverable)
    if discoverable:
        components = []
        component_data_list = _get_list(updates.get('components'), default=[])
        for component_data in component_data_list:
            components.append(parse_component_info(component_data, component))
        for component_data in components:
            if flask.current_app.config['FEDERATION_UUID'] in (component_data['uuid'], component_data['source_uuid']):
                # skip component info from or about this instance
                continue
            import_component_info(component_data, component)


def update_users(
        component: Component,
        updates: typing.Dict[str, typing.Any]
) -> None:
    _get_dict(updates, mandatory=True)
    _validate_header(updates.get('header'), component)

    users = []
    user_data_list = _get_list(updates.get('users'), default=[])
    user_email_hashes = _get_list(updates.get('federation_candidates'), default=[])
    for user_data in user_data_list:
        users.append(parse_user(user_data, component))
    for user_data in users:
        import_user(user_data, component)

    link_users_by_email_hashes(component.id, user_email_hashes)


def update_shares(
        component: Component,
        updates: typing.Dict[str, typing.Any]
) -> None:
    # parse and validate
    _get_dict(updates, mandatory=True)
    _validate_header(updates.get('header'), component)

    markdown_images = []
    markdown_images_data_dict = _get_dict(updates.get('markdown_images'), default={})
    for markdown_image_data in markdown_images_data_dict.items():
        markdown_images.append(parse_markdown_image(markdown_image_data, component))

    users = []
    user_data_list = _get_list(updates.get('users'), default=[])
    for user_data in user_data_list:
        users.append(parse_user(user_data, component))

    instruments = []
    instrument_data_list = _get_list(updates.get('instruments'), default=[])
    for instrument_data in instrument_data_list:
        instruments.append(parse_instrument(instrument_data))

    action_types = []
    action_type_data_list = _get_list(updates.get('action_types'), default=[])
    for action_type_data in action_type_data_list:
        action_types.append(parse_action_type(action_type_data))

    actions = []
    action_data_list = _get_list(updates.get('actions'), default=[])
    for action_data in action_data_list:
        actions.append(parse_action(action_data))

    location_types = []
    location_type_data_list = _get_list(updates.get('location_types'), default=[])
    for location_type_data in location_type_data_list:
        location_types.append(parse_location_type(location_type_data))

    locations = {}
    location_data_list = _get_list(updates.get('locations'), default=[])
    for location_data in location_data_list:
        location = parse_location(location_data)
        locations[(location['fed_id'], location['component_uuid'])] = location

    locations_check_for_cyclic_dependencies(locations)

    objects = []
    object_data_list = _get_list(updates.get('objects'), default=[])
    for object_data in object_data_list:
        objects.append(parse_object(object_data, component))

    # apply
    for markdown_image_data in markdown_images:
        import_markdown_image(markdown_image_data, component)

    for user_data in users:
        import_user(user_data, component)

    for instrument_data in instruments:
        import_instrument(instrument_data, component)

    for action_type_data in action_types:
        import_action_type(action_type_data, component)

    for action_data in actions:
        import_action(action_data, component)

    for location_type_data in location_types:
        import_location_type(location_type_data, component)

    while len(locations) > 0:
        key = list(locations)[0]
        import_location(locations[key], component, locations, users)
        if key in locations:
            locations.pop(key)

    import_status_by_object_id = {}
    for object_data in objects:
        import_status: typing.Dict[str, typing.Any] = {}
        import_object(object_data, component, import_status=import_status)
        import_status_by_object_id[object_data['fed_object_id']] = import_status

    if component.address:
        for object_id, import_status in import_status_by_object_id.items():
            if import_status:
                send_object_share_import_status(
                    object_id=object_id,
                    component=component,
                    import_status=import_status
                )


def send_object_share_import_status(
        object_id: int,
        component: Component,
        import_status: typing.Dict[str, typing.Any]
) -> None:
    put(
        endpoint=f'/federation/v1/shares/objects/{object_id}/import_status',
        component=component,
        json=import_status
    )



# coding: utf-8
"""
Script for creating an instrument in SampleDB.

Usage: sampledb create_instrument <name> <description>
"""

import sys
import typing

from .. import create_app
from ..logic.instruments import create_instrument, get_instruments
from ..logic.instrument_translations import set_instrument_translation
from ..logic.languages import Language


def main(arguments: typing.List[str]) -> None:
    if len(arguments) != 2:
        print(__doc__)
        sys.exit(1)
    name, description = arguments
    app = create_app()
    with app.app_context():
        if app.config['DISABLE_INSTRUMENTS']:
            print('Error: instruments are disabled', file=sys.stderr)
            sys.exit(1)
        instruments = get_instruments()
        for instrument in instruments:
            if instrument.name.get('en') == name:
                print(f'Error: an instrument with this name already exists (#{instrument.id})', file=sys.stderr)
                sys.exit(1)
        instrument = create_instrument()
        set_instrument_translation(
            language_id=Language.ENGLISH,
            instrument_id=instrument.id,
            name=name,
            description=description
        )
        print(f"Success: the instrument has been created in SampleDB (#{instrument.id})")



# coding: utf-8
"""
Add enable_object_assignments column to locations table and set it to True for existing locations.
"""

import flask_sqlalchemy

from .utils import table_has_column


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if table_has_column('locations', 'enable_object_assignments'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE locations
            ADD enable_object_assignments BOOLEAN DEFAULT TRUE NOT NULL
    """))
    return True
